import unittest
import logging
import IPython

class RBTree_Tests(unittest.TestCase):

    def setUp(self):
        self.n = Node(2)

    def tearDown(self):
        self.n = None

    def test_creation(self):
        self.assertIsNotNone(self.n)
        self.assertIsInstance(self.n, Node)
        self.assertIsNone(self.n.left)
        self.assertIsNone(self.n.right)
        self.assertIsNone(self.n.parent)
        self.assertTrue(self.n.red)

    def test_eq(self):
        self.assertTrue(self.n == self.n)
        other = Node(2)
        self.assertFalse(self.n == other)
        other2 = Node(5)
        self.assertFalse(self.n == other2)

    def test_blackheight(self):
        self.assertEqual(self.n.get_black_height(), 0)

    def test_blackheight_2(self):
        n1 = Node(1)
        n2 = Node(2, n1)
        n3 = Node(3, n2)
        self.assertEqual(n3.get_black_height(), 0)
        n3.red = False
        self.assertEqual(n3.get_black_height(), 1)
        n2.red = False
        self.assertEqual(n3.get_black_height(), 2)
        n1.red = False
        self.assertEqual(n3.get_black_height(), 3)

    def test_link_left(self):
        n1 = Node(2)
        n2 = Node(3)
        self.assertIsNone(n1.left)
        self.assertIsNone(n1.right)
        n1.link_left(n2)
        self.assertEqual(n1.left, n2)
        self.assertIsNone(n1.right)
        self.assertEqual(n2.parent, n1)

    def test_link_right(self):
        n1 = Node(2)
        n2 = Node(3)
        self.assertIsNone(n1.left)
        self.assertIsNone(n1.right)
        n1.link_right(n2)
        self.assertEqual(n1.right, n2)
        self.assertIsNone(n1.left)
        self.assertEqual(n2.parent, n1)

    def test_link_left_fails(self):
        n1 = Node(2)
        with self.assertRaises(Exception):
            n1.link_left(n1)
        n2 = Node(5)
        n1.link_left(n2)
        with self.assertRaises(Exception):
            n2.link_left(n1)

    def test_link_right_fails(self):
        n1 = Node(5)
        with self.assertRaises(Exception):
            n1.link_right(n1)
        n2 = Node(10)
        n1.link_right(n2)
        with self.assertRaises(Exception):
            n2.link_right(n1)

    def test_add_left(self):
        n1 = Node(2)
        n2 = Node(3)
        self.assertIsNone(n1.left)
        n1.add_left(n2)
        self.assertEqual(n1.left, n2)

    def test_add_right(self):
        n1 = Node(2)
        n2 = Node(3)
        self.assertIsNone(n1.right)
        n1.add_right(n2)
        self.assertEqual(n1.right, n2)

    def test_get_predecessor_None(self):
        n1 = Node(2)
        self.assertIsNone(n1.get_predecessor())

    def test_get_predecessor_basic(self):
        n1 = Node(2)
        n2 = Node(3)
        n1.add_left(n2)
        self.assertEqual(n1.get_predecessor(), n2)
        self.assertIsNone(n2.get_predecessor())

    def test_get_predecessor_basic_2(self):
        n1 = Node(2)
        n2 = Node(3)
        n1.add_right(n2)
        self.assertEqual(n2.get_predecessor(), n1)

    def test_get_predecessor(self):
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n1.add_right(n2)
        n2.add_left(n3)
        self.assertEqual(n3.get_predecessor(), n1)

    def test_get_successor_None(self):
        n1 = Node(2)
        self.assertIsNone(n1.get_successor())

    def test_get_successor_basic(self):
        n1 = Node(2)
        n2 = Node(3)
        n1.add_left(n2)
        self.assertEqual(n2.get_successor(), n1)

    def test_get_successor_basic_2(self):
        n1 = Node(2)
        n2 = Node(3)
        n1.add_right(n2)
        self.assertEqual(n1.get_successor(), n2)

    def test_get_successor(self):
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n1.add_left(n2)
        n2.add_right(n3)
        self.assertEqual(n3.get_successor(), n1)

    #test getPred/Succ_while

    def test_min(self):
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n1.add_left(n2)
        n2.add_left(n3)
        self.assertEqual(n1.min(), n3)

    def test_max(self):
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n1.add_right(n2)
        n2.add_right(n3)
        self.assertEqual(n1.max(), n3)

    def test_disconnect_from_parent(self):
        n1 = Node(2)
        n2 = Node(3)
        n1.add_left(n2)
        self.assertEqual(n2.parent, n1)
        self.assertEqual(n1.left, n2)
        n2.disconnect_from_parent()
        self.assertIsNone(n2.parent)
        self.assertIsNone(n1.left)

    def test_disconnect_left(self):
        n1 = Node(2)
        n2 = Node(3)
        n1.add_left(n2)
        self.assertEqual(n2.parent, n1)
        self.assertEqual(n1.left, n2)
        n1.disconnect_left()
        self.assertIsNone(n2.parent)
        self.assertIsNone(n1.left)

    def test_disconnect_right(self):
        n1 = Node(2)
        n2 = Node(3)
        n1.add_right(n2)
        self.assertEqual(n2.parent, n1)
        self.assertEqual(n1.right, n2)
        n1.disconnect_right()
        self.assertIsNone(n2.parent)
        self.assertIsNone(n1.right)

    def test_on_left(self):
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n1.add_left(n2)
        n1.add_right(n3)
        self.assertTrue(n1.on_left(n2))
        self.assertFalse(n1.on_left(n3))

    def test_rotate_right(self):
        n0 = Node(1)
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n4 = Node(5)
        n0.add_left(n1)
        n1.add_left(n2)
        n1.add_right(n3)
        n2.add_right(n4)
        setAsRoot, newHead = n1.rotate_right()
        #self.assertTrue(setAsRoot)
        self.assertEqual(newHead, n2)
        self.assertEqual(n1.left, n4)
        self.assertEqual(n4.parent, n1)
        self.assertEqual(n0.left, newHead)

    def test_rotate_left(self):
        n0 = Node(1)
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n4 = Node(5)
        n0.add_right(n1)
        n1.add_left(n2)
        n1.add_right(n3)
        n3.add_left(n4)
        setAsRoot, newHead = n1.rotate_left()
        #self.assertTrue(setAsRoot)
        self.assertEqual(newHead, n3)
        self.assertEqual(n1.right, n4)
        self.assertEqual(n4.parent, n1)
        self.assertEqual(n0.right, newHead)

    def test_get_pred_succ_while_no_condition(self):
        n1 = Node(2)
        n2 = Node(3)
        n3 = Node(4)
        n4 = Node(5)
        n1.add_left(n2)
        n1.add_right(n3)
        n3.add_left(n4)
        succresults = n2.get_successor_while(lambda x: True)
        self.assertEqual(succresults, [n1,n4,n3])
        predresults = n3.get_predecessor_while(lambda x: True)
        self.assertEqual(predresults, [n4,n1,n2])
        neighbours = n1.get_neighbours_while(lambda x: True)
        self.assertEqual(neighbours, [n2,n4,n3])

    def test_is_leaf(self):
        n1 = Node(2)
        self.assertTrue(n1.is_leaf())
        n1.add_left(Node(34))
        self.assertFalse(n1.is_leaf())
        n1.add_right(Node(5))
        self.assertFalse(n1.is_leaf())

    #----------

if __name__ == "__main__":
    #use python $filename to use this logging setup
      LOGLEVEL = logging.INFO
      logFileName = "log.rbtree_tests"
      logging.basicConfig(filename=logFileName, level=LOGLEVEL, filemode='w')
      console = logging.StreamHandler()
      console.setLevel(logging.WARN)
      logging.getLogger().addHandler(console)
      unittest.main()
      #reminder: user logging.getLogger().setLevel(logging.NOTSET) for log control



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class VoronoiProcess:

    def finalise_dcel(self, constrain_to_bbox=True, radius=100):
        """ Cleanup the DCEL of the voronoi diagram,
            completing faces and constraining to a bbox
        """
        if bool(self.events):
            logging.warninging("Finalising with events still to process")
        logging.debug("-------------------- Finalising DCEL")
        logging.debug(self.dcel)

        self._update_arcs(self.sweep_position.y() - 1000)
        #Not a pure DCEL operation as it requires curve intersection:
        self._complete_edges()
        self.dcel.purge()
        tempbbox = self.bbox + np.array([100, 100, -100, -100])
        #np.array([100, 100, -100, -100])
        if constrain_to_bbox:
            #modify or mark edges outside bbox
            self.dcel.constrain_to_bbox(tempbbox, force=True)
        else:
            centre = bbox_centre(self.bbox)
            self.dcel.constrain_to_circle(centre, radius)
        self.dcel.purge()
        logging.debug("---------- Constrained to bbox")
        #ensure CCW ordering
        for f in self.dcel.faces:
            f.fixup(tempbbox)
        #cleanup faces
        logging.debug("---------- Fixed up faces")
        self.dcel.purge()
        logging.debug("---------- Purged 3")
        logging.debug(self.dcel)
        self.dcel.verify_all()
        return self.dcel

    def calculate_to_completion(self):
        """ Calculate the entire voronoi for all points """
        finished = False
        #Max Steps for a guaranteed exit
        while not finished and self.current_step < self.max_steps:
            logging.debug("----------------------------------------")
            logging.debug("Calculating step: {}".format(self.current_step))
            finished = self._calculate()
            if self.debug_draw:
                self.debug.draw_intermediate_states(self.current_step, dcel=True, text=True)
            self.current_step += 1

    def _calculate(self):
        """ Calculate the next step of the voronoi diagram,
            Return True on completion, False otherwise
        """
        if not bool(self.events): #finished calculating, early exit
            return True
        ##Get the next event
        event = heapq.heappop(self.events)
        #update the sweep position
        self.sweep_position = event
        logging.debug("Sweep position: {}".format(self.sweep_position.loc))
        #update the arcs:
        self._update_arcs(self.sweep_position.y())
        #handle the event:
        if isinstance(event, SiteEvent):
            self._handle_site_event(event)
        elif isinstance(event, CircleEvent):
            if event.active:
                self._handle_circle_event(event)
            else:
                logging.debug("-------------------- Skipping deactivated circle event")
                logging.debug(event)
        else:
            raise Exception("Unrecognised Event")
        return False

    def _calculate_circle_events(self, node, left=True, right=True):
        """
        Given an arc node, get the arcs either side, and determine if/when it will disappear
        """
        logging.debug("Calculating circle events for: {}".format(node))
        #Generate a circle event for left side, and right side
        left_triple = self.beachline.get_predecessor_triple(node)
        right_triple = self.beachline.get_successor_triple(node)
        #Calculate chords and determine circle event point:
        #add circle event to events and the relevant leaf
        if left_triple:
            logging.debug("Calc Left Triple: {}".format("-".join([str(x) for x in left_triple])))

        if left and left_triple and left_triple[0].value != left_triple[2].value:
            left_points = np.array([x.value.get_focus() for x in left_triple])
            left_circle = utils.math.get_circle_3p(*left_points)

            #possibly use ccw for this, with glpoc from below
            if left_circle is not None and isClockwise(*left_points):
                left_circle_loc = utils.math.get_lowest_point_on_circle(*left_circle)
                #check the l_t_p/s arent in this circle
                #note: swapped this to add on the right ftm
                self._add_circle_event(left_circle_loc, left_triple[1], left_circle[0], left=True)
            else:
                logging.debug("Left points failed: {}".format(left_points))

        if right_triple:
            logging.debug("Calc Right Triple: {}".format("-".join([str(x) for x in right_triple])))

        if right and right_triple and right_triple[0].value != right_triple[2].value:
            right_points = np.array([x.value.get_focus() for x in right_triple])
            right_circle = utils.math.get_circle_3p(*right_points)
            if right_circle is not None and isClockwise(*right_points):
                right_circle_loc = utils.math.get_lowest_point_on_circle(*right_circle)
                #note: swapped this to add on the left ftm
                self._add_circle_event(right_circle_loc,
                                       right_triple[1],
                                       right_circle[0],
                                       left=False)
            else:
                logging.debug("Right points failed: {}".format(right_points))

    def _update_arcs(self, d):
        """ Trigger the update of all stored arcs with a new frontier line position """
        self.beachline.update_values(lambda v, q: v.update_d(q), d)

    #-------------------- DCEL Completion


    def _add_circle_event(self, loc, source_node, voronoi_vertex, left=True):
        if loc[1] > self.sweep_position.y():# or np.allclose(loc[1], self.sweep_position.y()):
            logging.debug("Breaking out of add circle event: Wrong side of Beachline")
            return
        event = CircleEvent(loc, source_node, voronoi_vertex, i=self.current_step, left=left)
        logging.debug("Adding: {}".format(event))
        heapq.heappush(self.events, event)
        self.circles.append(event)

    def _delete_circle_events(self, node, pre=None, post=None, event=None):
        """ Deactivate a circle event rather than deleting it.
        This means instead of removal and re-heapifying, you just skip the event
        when you come to process it """
        logging.debug("Deactivating Circle Event: {}".format(node))
        if node is not None:
            if CIRCLE_EVENTS.LEFT in node.data:
                node.data[CIRCLE_EVENTS.LEFT].deactivate()
            if CIRCLE_EVENTS.RIGHT in node.data:
                node.data[CIRCLE_EVENTS.RIGHT].deactivate()

        if pre is not None and CIRCLE_EVENTS.RIGHT in pre.data:
            pre.data[CIRCLE_EVENTS.RIGHT].deactivate()
        if post is not None and CIRCLE_EVENTS.LEFT in post.data:
            post.data[CIRCLE_EVENTS.LEFT].deactivate()



    def relax(self, amnt=0.5, faces=None):
        """ Having calculated the voronoi diagram, use the centroids of
            the faces instead of the sites, and rerun the calculation.
        Can be passed in a subset of faces
        """
        assert(not bool(self.events))

        if faces is None:
            faces = self.dcel.faces
        #Figure out any faces that are excluded
        face_indices = set([x.index for x in faces])
        other_face_sites = np.array([x.site for x in self.dcel.faces
                                     if x.index not in face_indices])
        #Get a line of origin - centroid
        lines = np.array([np.concatenate((x.site, x.getAvgCentroid())) for x in faces])
        #Move along that line toward the centroid
        new_sites = np.array([utils.math.sample_along_lines(*x, amnt)[0] for x in lines])
        #Combine with excluded faces
        if bool(other_face_sites) and bool(new_sites):
            total_sites = np.row_stack((new_sites, other_face_sites))
        elif bool(new_sites):
            total_sites = new_sites
        else:
            total_sites = other_face_sites
        assert(len(self.dcel.faces) == len(total_sites))
        #Setup the datastructures with the new sites
        self.reset()
        self.init_graph(data=new_sites, rerun=True)
        self.calculate_to_completion()


class _VoronoiHandlers:

    def _handle_site_event(self, event):
        """
        provided with a site event, add it to the beachline in the appropriate place
        then update/remove any circle events that trios of arcs generate
        """
        assert(isinstance(event, SiteEvent))
        logging.debug("Handling Site Event: {}".format(event))
        #The new parabola made from the site
        new_arc = Parabola(*event.loc, self.sweep_position.y())
        #get the x position of the event
        x_pos = new_arc.fx

        #if beachline is empty: add and return
        if not bool(self.beachline):
            new_node = self.beachline.insert(new_arc)[0]
            new_node.data['face'] = event.face
            return

        #Otherwise, slot the arc between existing nodes
        closest_node, direction = self._get_closest_arc_node(x_pos)
        assert(closest_node is not None)
        #remove the obsolete circle event
        self._delete_circle_events(closest_node)
        new_node, duplicate_node = self._split_beachline(direction,
                                                         closest_node,
                                                         new_arc,
                                                         event.face)

        #Create an edge between the two nodes, without origin points yet
        logging.debug("Adding edge on side: {}".format(direction))
        node_face = closest_node.data['face']
        if direction is Directions.LEFT:
            the_face = event.face
            twin_face = node_face
            node_pair = (new_node, closest_node)
        else:
            the_face = node_face
            twin_face = event.face
            node_pair = (closest_node, new_node)

        new_edge = self.dcel.new_edge(None, None, face=the_face, twin_face=twin_face,
                                      edata=BASE_VORONOI_EDGE_DATA, vdata=BASE_VORONOI_VERT_DATA)
        self._store_edge(new_edge, *node_pair)
        self._cleanup_edges(direction, new_edge, new_node, closest_node, duplicate_node)

        #create circle events:
        self._calculate_circle_events(new_node)

    def _handle_circle_event(self, event):
        """
        provided a circle event, add a new vertex to the dcel,
        then update the beachline to connect the two sides of the arc that has disappeared
        """
        assert(isinstance(event, CircleEvent))
        logging.debug("Handling Circle Event: {}".format(event))
        #remove disappearing arc from tree
        #and update breakpoints, remove false alarm circle events
        node = event.source
        pre = node.getPredecessor()
        suc = node.getSuccessor()
        assert('face' in pre.data)
        assert('face' in suc.data)

        self._delete_circle_events(node, pre, suc, event)

        #add the centre of the circle causing the event as a vertex record
        logging.debug("Creating Vertex")
        new_vertex = self.dcel.new_vertex(event.vertex, data=BASE_VORONOI_VERT_DATA)

        #attach the vertex as a defined point in the half edges for the three faces,
        #these faces are pre<->node and node<->succ

        e1 = self._get_edge(pre, node)
        e2 = self._get_edge(node, suc)

        #create two half-edge records for the new breakpoint of the beachline
        logging.debug("Creating a new half-edge {}-{}".format(pre, suc))
        new_edge = self.dcel.new_edge(new_vertex, None,
                                      face=pre.data['face'],
                                      twin_face=suc.data['face'],
                                      edata=BASE_VORONOI_EDGE_DATA,
                                      vdata=BASE_VORONOI_VERT_DATA)

        if e1:
            #predecessor face
            logging.debug("Adding vertex to {}-{}".format(pre, node))
            assert(e1.face == pre.data['face'])
            assert(e1.twin.face == node.data['face'])
            e1.addVertex(new_vertex)
            e1.addPrev(new_edge, force=True)
        else:
            logging.debug("No r-edge found for {}-{}".format(pre, node))

        if e2:
            #successor face
            logging.debug("Adding vertex to {}-{}".format(node, suc))
            assert(e2.twin.face == suc.data['face'])
            assert(e2.face == node.data['face'])
            e2.addVertex(new_vertex)
            e2.twin.addNext(new_edge.twin, force=True)
        else:
            logging.debug("No r-edge found for {}-{}".format(node, suc))

        #store the new edge, but only for the open breakpoint
        #the breakpoint being the predecessor and successor, now partners following
        #removal of the middle node above in this function
        self._store_edge(new_edge, pre, suc)

        #delete the node, no longer needed as the arc has reduced to 0
        logging.debug("Pre-Deletion: {}".format(self.beachline.get_chain()))
        self.beachline.delete(node)
        logging.debug("Post-Deletion: {}".format(self.beachline.get_chain()))
        #recheck for new circle events
        if pre:
            self._calculate_circle_events(pre, left=False, right=True)
        if suc:
            self._calculate_circle_events(suc, left=True, right=False)

class _VoronoiUtils:

    def _cleanup_edges(self, direction, edge, new_node, node, duplicate_node):
        """ if there was an edge of closest_arc -> closest_arc.successor: update it
        because closest_arc is not adjacent to successor any more, duplicate_node is """
        if direction is Directions.LEFT:
            logging.debug("Cleaning up left")
            dup_node_sibling = duplicate_node.getPredecessor()
            if dup_node_sibling is not None:
                e1 = self._get_edge(dup_node_sibling, node)
                if e1 is not None:
                    self._remove_edge(dup_node_sibling, node)
                    self._store_edge(e1, dup_node_sibling, duplicate_node)
        else:
            logging.debug("Cleaning up right")
            dup_node_sibling = duplicate_node.getSuccessor()
            if dup_node_sibling is not None:
                e1 = self._get_edge(node, dup_node_sibling)
                if e1 is not None:
                    self._remove_edge(node, dup_node_sibling)
                    self._store_edge(e1, duplicate_node, dup_node_sibling)

        if direction is Directions.LEFT:
            self._store_edge(edge, new_node, node)
            self._store_edge(edge.twin, duplicate_node, new_node)
        else:
            self._store_edge(edge, node, new_node)
            self._store_edge(edge.twin, new_node, duplicate_node)

    def _get_closest_arc_node(self, x_pos):
        #search for the breakpoint interval of the beachline
        closest_arc_node, direction = self.beachline.search(x_pos, closest=True)
        if closest_arc_node is not None:
            logging.debug("Closest Arc Triple: {} *{}* {}".format(closest_arc_node.getPredecessor(),
                                                                  closest_arc_node,
                                                                  closest_arc_node.getSuccessor()))
            logging.debug("Direction: {}".format(direction))
        return (closest_arc_node, direction)

    def _split_beachline(self, direction, node, arc, event_face):
        #If site is directly below the arc, or on the right of the arc, add it as a successor
        if direction is Directions.RIGHT:
            new_node = self.beachline.insert_successor(node, arc)
            duplicate_node = self.beachline.insert_successor(new_node, node.value)
        else:
            #otherwise add it as a predecessor
            new_node = self.beachline.insert_predecessor(node, arc)
            duplicate_node = self.beachline.insert_predecessor(new_node, node.value)
        assert(isinstance(new_node, rbtree.Node))

        #add in the faces as a data point for the new node and duplicate
        new_node.data['face'] = event_face
        duplicate_node.data['face'] = node.data['face']

        #Debug the new triple: [ A, B, A]
        triple_string = "-".join([repr(x) for x in [node, new_node, duplicate_node]])
        logging.debug("Split {} into {}".format(repr(node), triple_string))
        return new_node, duplicate_node

    #-------------------- Fortune Methods

    def _complete_edges(self):
        """ get any infinite edges, and complete their intersections """
        logging.debug("\n---------- Infinite Edges Completion")
        i = 0

        #get only the halfedges that are originless, rather than full edges that are infinite
        i_pairs = [x for x in self.halfedges.items() if x[1].isInfinite()]
        logging.debug("Origin-less half edges num: {}".format(len(i_pairs)))

        #----
        #i_pairs = [((breakpoint nodes), edge)]
        for (bw, c) in i_pairs:
            i += 1
            #a and b are nodes
            a = bw.bp1
            b = bw.bp2
            debug_string = "{} Infinite Edge resolution: {}-{}, infinite? {}"
            logging.debug(debug_string.format(i, a, b, c.isInfinite()))
            if c.origin is None and c.twin.origin is None:
                logging.debug("Found an undefined edge, cleaning up")
                c.markForCleanup()
                continue
            if not c.isInfinite():
                continue
            #raise Exception("An Edge is not infinite")
            #intersect the breakpoints to find the vertex point
            intersection = a.value.intersect(b.value)
            if intersection is None or not bool(intersection):
                raise Exception("No intersections detected when completing an infinite edge")
            elif len(intersection) == 2:
                verts = [x for x in c.getVertices() if x is not None]
                assert(len(verts) == 1)
                lines = []
                lines += bound_line_in_bbox(np.array([verts[0].toArray(), intersection[0]]),
                                            self.bbox)
                lines += bound_line_in_bbox(np.array([verts[0].toArray(), intersection[1]]),
                                            self.bbox)
                distances = np.array([get_distance_raw(x[:2], x[2:]) for x in lines])
                min_line = np.argmin(distances)
                new_vertex = self.dcel.new_vertex(lines[min_line][1], data=BASE_VORONOI_VERT_DATA)
                c.addVertex(new_vertex)

            if c.isInfinite():
                logging.debug("Edge is still infinite, marking for cleanup")
                c.markForCleanup()

    #-------------------- Beachline Edge Interaction

    def _store_edge(self, edge, bp1, bp2):
        """ Store an incomplete edge by the 2 pairs of nodes that define the breakpoints """
        assert(isinstance(edge, HalfEdge))
        assert(isinstance(bp1, rbtree.Node))
        assert(isinstance(bp2, rbtree.Node))
        if self._has_edge(bp1, bp2) and self._get_edge(bp1, bp2) != edge:
            raise Exception("Overrighting edge breakpoint: {}, {}".format(bp1, bp2))
        logging.debug("Storing Edge: ({}, {}): {}".format(bp1, bp2, edge))
        self.halfedges[BreakWrapper(bp1, bp2)] = edge

    def _has_edge(self, bp1, bp2):
        assert(bp1 is None or isinstance(bp1, rbtree.Node))
        assert(bp2 is None or isinstance(bp2, rbtree.Node))
        return BreakWrapper(bp1, bp2) in self.halfedges

    def _get_edge(self, bp1, bp2):
        assert(bp1 is None or isinstance(bp1, rbtree.Node))
        assert(bp2 is None or isinstance(bp2, rbtree.Node))
        if self._has_edge(bp1, bp2):
            return self.halfedges[BreakWrapper(bp1, bp2)]
        else:
            return None

    def _remove_edge(self, bp1, bp2):
        assert(isinstance(bp1, rbtree.Node))
        assert(isinstance(bp2, rbtree.Node))
        if not self._has_edge(bp1, bp2):
            raise Exception("trying to remove a non-existing edge")
        logging.debug("Removing Edge: ({}, {}) : {}".format(bp1,
                                                            bp2,
                                                            self.halfedges[BreakWrapper(bp1, bp2)]))
        del self.halfedges[BreakWrapper(bp1, bp2)]

    #-------------------- Circle Event Interaction



"""
A Basic Tree Module
"""
##-- imports
from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

from functools import partial
from string import ascii_uppercase
from types import FunctionType
import logging as root_logger
##-- end imports

logging = root_logger.getLogger(__name__)

@dataclass
class BinaryTree:
    """
    A Simple Binary Tree Class. Trees are nodes themselves.
    Each node can hold a *value* (used for searching, ordering etc),
    and *data*, that can be anything, especially a dictionary.
    """

    value  : Any  = field()
    data   : Any  = field(default=None)
    left   : Tree = field(default=None)
    right  : Tree = field(default=None)
    parent : Tree = field(default=None)
    root   : bool = field(default=False)

    i = 0

    def __post_init__(self):
        self.id = Tree.i
        Tree.i += 1


    def __str__(self):
        if self.left is not None:
            left_string = self.left.__str__()
        else:
            left_string = "()"
        if self.right is not None:
            right_string = self.right.__str__()
        else:
            right_string = "()"
        return "( V: {} Left: {}, Right: {} )".format(self.value, left_string, right_string)

    def __repr__(self):
        if self.left is not None:
            left_string = self.left.__str__()
        else:
            left_string = "()"
        if self.right is not None:
            right_string = self.right.__str__()
        else:
            right_string = "()"
        return "<TreeNode>"

    def __hash__(self):
        if self.value is not None and hasattr(self.value, "id"):
            return self.value.id
        return self.id
    def __eq__(self, other):
        assert(other is None or isinstance(other, Tree))
        result = False
        if other is None:
            return result
        result = self.id == other.id
        return result

    def is_leaf(self):
        """ Test if this object is a leaf node """
        return self.left is None and self.right is None

    def is_root(self):
        """ Test if this object is the root of a tree """
        return self.root

    def insert(self, value, data=None):
        """ Insert a node into the tree """
        insert_on_left = value < self.value
        if insert_on_left:
            if self.left is None:
                self.left = Tree(value, data=data)
            else:
                self.left.insert(value, data=data)
        else:
            if self.right is None:
                self.right = Tree(value, data=data)
            else:
                self.right.insert(value, data=data)

    def search(self, value) -> Optional[Any]:
        """ Search the tree for a value """
        if value == self.value:
            return self
        elif self.is_leaf():
            return None
        elif value < self.value:
            if self.left is None:
                return None
            else:
                return self.left.search(value)
        else:
            if self.right is None:
                return None
            else:
                return self.right.search(value)

    def get_range(self, l, r):
        """ Get the leftmost and rightmost values of the tree """
        values = []
        if l < self.value and self.left is not None:
            values.extend(self.left.get_range(l, r))

        if l < self.value and self.value <= r:
            values.append(self)

        if self.value <= r and self.right is not None:
            values.extend(self.right.get_range(l, r))

        return values



    def min(self):
        """ Get the leftmost node of the tree """
        current = self
        while current.left is not None:
            current = current.left
        return current

    def max(self):
        """ Get the rightmost node of the tree """
        current = self
        while current.right is not None:
            current = current.right
        return current

    def get_predecessor(self):
        """ Get the node to the immediate left """
        if self.left is not None:
            return self.left.max()
        if self.parent is not None and not self.parent.on_left(self):
            return self.parent
        prev = self
        current = self.parent
        count = 0
        while current is not None and current.right != prev:
            prev = current
            current = current.parent
            count += 1

        if current is not self:
            return current
        else:
            return None

    def get_successor(self):
        """ Get the node to the immediate right """
        if self.right is not None:
            return self.right.min()
        if self.parent is not None and self.parent.on_left(self):
            return self.parent
        prev = self
        current = self.parent
        while current is not None and current.left != prev:
            prev = current
            current = current.parent

        if current is not self:
            return current
        else:
            return None

    def get_predecessor_while(self, condition):
        """ Collect predecessors while the condition is true """
        assert(isinstance(condition, (FunctionType, partial)))
        results = []
        current = self.get_predecessor()
        while current is not None and condition(current):
            results.append(current)
            current = current.get_predecessor()
        return results

    def get_successor_while(self, condition):
        """ Collect successors while the condition is true """
        assert(isinstance(condition, (FunctionType, partial)))
        results = []
        current = self.get_successor()
        while current is not None and condition(current):
            results.append(current)
            current = current.get_successor()
        return results

    def get_neighbours_while(self, condition):
        """ Collect left and right nodes whlie the condition is true """
        results = []
        results += self.get_predecessor_while(condition)
        results += self.get_successor_while(condition)
        return results

    #------------------------------
    # def Basic Update
    #------------------------------
    def add_left(self, node, force=False):
        """ Add a node to the immediate left """
        if self == node:
            node = None
        if self.left is None or force:
            self.link_left(node)
        else:
            self.get_predecessor().add_right(node)
        logging.debug("{}: Adding {} to Left".format(self, node))

    def add_right(self, node, force=False):
        """ Add a node to the immediate right """
        if self == node:
            node = None
        if self.right is None or force:
            self.link_right(node)
        else:
            self.get_successor().add_left(node)
        logging.debug("{}: Adding {} to Right".format(self, node))

    def link_left(self, node):
        """ Connect the passed in node as a predecessor """
        assert(node is not self)
        if node is not None:
            assert(self.right is not node)
            assert(self.parent is not node)
            assert(node.left is not self)
            assert(node.right is not self)
        self.left = node
        if self.left is not None:
            self.left.parent = self
        logging.debug("{} L-> {}".format(self, node))

    def link_right(self, node):
        """ Connect the passed in node as a successor """
        assert(node is not self)
        if node is not None:
            assert(self.parent is not node)
            assert(node.left is not self)
            assert(node.right is not self)
            assert(self.left is not node)
        self.right = node
        if self.right is not None:
            self.right.parent = self
        logging.debug("{} R-> {}".format(self, node))

    def disconnect_from_parent(self):
        """ Symmetrically disconnect this node from its parent """
        parent = self.parent
        if self.parent is not None:
            if self.parent.on_left(self):
                self.parent.left = None
            else:
                self.parent.right = None
            self.parent = None
        logging.debug("Disconnecting {} -> {}".format(parent, self))

    def disconnect_left(self):
        """ Symmetrically disconnect this node from is predecessor """
        if self.left is not None:
            node = self.left
            self.left = None
            node.parent = None
            logging.debug("{} disconnecting left".format(self))
            return node
        return None

    def disconnect_right(self):
        """ Symmetrically disconnect this node from is successor """
        if self.right is not None:
            node = self.right
            self.right = None
            node.parent = None
            logging.debug("{} disconnecting right".format(self))
            return node
        return None

    def on_left(self, node):
        """ Test to see if the passed in node is the predecessor """
        assert(isinstance(node, Tree))
        return self.left == node

    def rotate_right(self):
        """ Trigger a rotation to the right centered on this node """
        set_as_root = True
        orig_parent = None
        originally_on_left = False
        new_head = self.left
        new_right = self
        new_left = new_head.right
        if self.parent is not None:
            set_as_root = False
            originally_on_left = self.parent.on_left(self)
            orig_parent = self.parent
            new_right.disconnect_from_parent()
        new_head.disconnect_from_parent()
        if new_left is not None:
            new_left.disconnect_from_parent()

        new_right.link_left(new_left)
        new_head.link_right(new_right)
        if orig_parent is not None:
            if originally_on_left:
                orig_parent.link_left(new_head)
            else:
                orig_parent.link_right(new_head)
        return set_as_root, new_head

    def rotate_left(self):
        """ Trigger a rotation to the left centered on this node """
        set_as_root = True
        orig_parent = None
        originally_on_left = False
        new_head = self.right
        new_left = self
        new_right = new_head.left
        if self.parent is not None:
            set_as_root = False
            originally_on_left = self.parent.on_left(self)
            orig_parent = self.parent
            new_left.disconnect_from_parent()
        new_head.disconnect_from_parent()
        if new_right is not None:
            new_right.disconnect_from_parent()

        new_left.link_right(new_right)
        new_head.link_left(new_left)
        if orig_parent is not None:
            if originally_on_left:
                orig_parent.link_left(new_head)
            else:
                orig_parent.link_right(new_head)
        return set_as_root, new_head



#/usr/bin/env python3
"""
AST representations bridging parsed DSL -> compiled clingo
"""
##-- imports
from __future__ import annotations

import logging as logmod
import pathlib as pl
from os import getcwd
from enum import Enum, auto
from dataclasses import InitVar, dataclass, field
import re
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref
##-- end imports

logging = logmod.getLogger(__name__)

VAR_SIG_REG = re.compile(r"\d+$")

##-- util context manager

class ASTContextManager:
    """ For ensuring all ASTs are built with the correct source """

    def __init__(self, parse_source):
        self.parse_source = parse_source

    def __enter__(self):
        SolverAST.current_parse_source = self.parse_source

    def __exit__(self, exc_type, exc_value, exc_traceback):
        SolverAST.current_parse_source = None

##-- end util context manager

@dataclass(frozen=True)
class SolverAST:
    parse_source : list[str|pl.Path]    = field(default_factory=list, kw_only=True, repr=False)
    parse_loc    : None|tuple[int, int] = field(default=None, kw_only=True)

    current_parse_source : ClassVar[None|str] = None

    def __post_init__(self):
        if SolverAST.current_parse_source is not None:
            self.parse_source.append(SolverAST.current_parse_source)

    @property
    def sources_str(self):
        if not bool(self.parse_source):
            return "n/a"

        full_path = self.parse_source[0]
        cwd       = getcwd()
        match full_path:
            case pl.Path():
                return str(full_path.relative_to(cwd))
            case str():
                return full_path
            case _:
                raise TypeError("An AST has an unexpected pasrse source", full_path)

    @staticmethod
    def manage_source(parse_source):
        return ASTContextManager(parse_source)

@dataclass(frozen=True)
class TermAST(SolverAST):
    value  : str             = field()
    params : list[TermAST]   = field(default_factory=list)
    is_var : bool            = field(default=False)

    def __post_init__(self):
        assert(not (self.is_var and bool(self.params)))
        super().__post_init__()

    def __str__(self):
        if bool(self.params):
            param_str = ",".join(str(x) for x in self.params)
            return self.value + "(" + param_str + ")"

        return str(self.value)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, TermAST):
            return False

        if not self.value == other.value:
            return False

        return all(x == y for x,y in zip(self.params, other.params))

    @property
    def signature(self):
        if not self.is_var:
            return f"{self.value}/{len(self.params)}"

        chopped = VAR_SIG_REG.sub("", self.value)
        return f"{chopped}/{len(self.params)}"

    @property
    def has_var(self) -> bool:
        return self.is_var or any(x.has_var for x in self.params)



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


class JGDVPathSimpleKey(JGDVSimpleKey):
    """ A Key that always expands as a path """

    def expand(self, spec=None, state=None, *, rec=False, insist=False, chain:list[JGDVBaseKey]=None, on_fail=Any, locs=None, **kwargs):
        return str(self.to_path(spec, state, chain=chain, on_fail=on_fail, locs=locs))

    def __repr__(self):
        return "<DootPathSimpleKey: {}>".format(str(self))

    def __call__(self, spec, state):
        """ Expand the key using the registered expansion hint """
        match getattr(self, EXPANSION_HINT, False):
            case False:
                return self.to_path(spec, state)
            case {"expansion": "str", "kwargs": kwargs}:
                return self.expand(spec, state, **kwargs)
            case {"expansion": "path", "kwargs": kwargs}:
                return self.to_path(spec, state, **kwargs)
            case {"expansion": "redirect"}:
                return self.redirect(spec)
            case {"expansion": "redirect_multi"}:
                return self.redirect_multi(spec)
            case x:
                raise doot.errors.DootKeyError("Key Called with Bad Key Expansion Type", self, x)


class JGDVPathMultiKey(JGDVMultiKey):
    """ A MultiKey that always expands as a path """

    def expand(self, spec=None, state=None, *, rec=False, insist=False, chain:list[JGDVBaseKey]=None, on_fail=Any, locs=None, **kwargs):
        return str(self.to_path(spec, state, chain=chain, on_fail=on_fail, locs=locs))

    def __repr__(self):
        return "<DootPathMultiKey: {}>".format(str(self))

    def __call__(self, spec, state):
        """ Expand the key using the registered expansion hint """
        match getattr(self, EXPANSION_HINT, False):
            case False:
                return self.to_path(spec, state)
            case {"expansion": "str", "kwargs": kwargs}:
                return self.expand(spec, state, **kwargs)
            case {"expansion": "path", "kwargs": kwargs}:
                return self.to_path(spec, state, **kwargs)
            case {"expansion": "redirect"}:
                return self.redirect(spec)
            case {"expansion": "redirect_multi"}:
                return self.redirect_multi(spec)
            case x:
                raise doot.errors.DootKeyError("Key Called with Bad Key Expansion Type", self, x)



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import os
import sys
import atexit
import faulthandler
import signal

env : dict = os.environ

def exception_hook(exc_type, exc_value, tb):
    """ A Basic exception handler example
      from https://python.plainenglish.io/creating-beautiful-tracebacks-with-pythons-exception-hooks-c8a79e13558d
    """
    print('Traceback:')
    filename = tb.tb_frame.f_code.co_filename
    name = tb.tb_frame.f_code.co_name
    line_no = tb.tb_lineno
    print(f"File {filename} line {line_no}, in {name}")

    # Exception type and value
    print(f"{exc_type.__name__}, Message: {exc_value}")

def exception_hook_loop(exc_type, exc_value, tb):
    ""
    local_vars = tb.tb_frame.f_locals
    tb = tb.tb_next
    print(f"Local variables in top frame: {local_vars}")

class JGDVHookConfig:
    """
    Utiility class for setting up various python hooks
    """

    def __init__(self):
        self._disabled = "PRE_COMMIT" in env
        self._exit_hook       = None
        self._exception_hook  = None
        self._breakpoint_hook = None
        self._display_hook    = None
        self._unraisable_hook = None

    def setup(self):
        pass

    def set_fault(self):
        faulthandler.enable(file=sys.stderr, all_threads=True)

    def set_exit(self):
        if not self._exit_hook:
            return

        @atexit.register(self._exit_hook)

    def set_exception(self):
        if not self._exception_hook:
            return
        sys.excepthook = self._exception_hook

    def set_breakpoint(self):
        if not self._breakpoint_hook:
            return
        sys.breakpointhook = self._breakpointhook

    def set_display(self):
        if not self._display_hook:
            reutrn
        sys.displayhook = self._display_hook

    def set_unraisable(self):
        if not self._unraisable_hook:
            return

        self.unraisablehook = self._unraisable_hook


    def set_signal(self, sig=signal.SIGINT):
        printer.debug("Installing Task Loop handler for: %s", signal.strsignal(sig))
        # Install handler for Interrupt signal
        signal.signal(sig, SignalHandler.handle)

    def unset_signal(self, sig=signal.SIGINT):
        printer.debug("Uninstalling Task Loop handler for: %s", signal.strsignal(sig))
        signal.signal(sig, signal.SIG_DFL)

    def __enter__(self):
        if not self._disabled:
            self.set_signal()
        return

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not self._disabled:
            self.unset_signal()
        # return False to reraise errors
        return



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv.key import JGDVKey
from jgdv.enums.location_meta import LocationMeta
from jgdv.enums.loop_control import LoopControl

MARKER : Final[str] = doot.constants.paths.MARKER_FILE_NAME
walk_ignores : Final[list] = doot.config.on_fail(['.git', '.DS_Store', "__pycache__"], list).settings.walking.ignores()
walk_halts   : Final[str]  = doot.config.on_fail([".doot_ignore"], list).settings.walking.halts()

class PathManip_m:
    """
      A Mixin for common path manipulations
    """

    def _calc_path_parts(self, fpath, roots) -> dict:
        """ take a path, and get a dict of bits to add to state from it """
        assert(fpath is not None)

        temp_stem  = fpath
        # This handles "a/b/c.tar.gz"
        while temp_stem.stem != temp_stem.with_suffix("").stem:
            temp_stem = temp_stem.with_suffix("")

        return {
            'rpath'   : self._get_relative(fpath, roots),
            'fstem'   : temp_stem.stem,
            'fparent' : fpath.parent,
            'fname'   : fpath.name,
            'fext'    : fpath.suffix,
            'pstem'   : fpath.parent.stem,
            }

    def _build_roots(self, spec, state, roots) -> list[pl.Path]:
        """
        convert roots from keys to paths
        """
        roots = roots or []
        results = []

        for root in roots:
            root_key = JGDVKey.build(root, explicit=True)
            results.append(root_key.to_path(spec, state))

        return results

    def _get_relative(self, fpath, roots) -> pl.Path:
        logging.debug("Finding Relative Path of: %s using %s", fpath, roots)
        if not fpath.is_absolute():
            return fpath

        if not bool(roots):
            return None

        for root_path in roots:
            try:
                return fpath.relative_to(root_path)
            except ValueError:
                continue

        raise ValueError(f"{fpath} is not able to be made relative", roots)

    def _shadow_path(self, rpath:pl.Path, shadow_root:pl.Path) -> pl.Path:
        """ take a relative path, apply it onto a root to create a shadowed location """
        assert(isinstance(rpath, pl.Path))
        assert(not rpath.is_absolute())
        result      = shadow_root / rpath
        if result == doot.locs[rpath]:
            raise doot.errors.DootLocationError("Shadowed Path is same as original", fpath)

        return result.parent

    def _find_parent_marker(self, fpath, marker=None) -> None|pl.Path:
        """ Go up the parent list to find a marker file, return the dir its in """
        marker = marker or MARKER
        for p in fpath.parents:
            if (p / marker).exists():
                return p

        return None

    def _is_write_protected(self, fpath) -> bool:
        for key in filter(lambda x: doot.locs.metacheck(x, LocationMeta.protected), doot.locs):
            base = getattr(doot.locs, key)
            if fpath.is_relative_to(base):
                return True

        return False

    def _normalize(self, path:pl.Path, root=None, symlinks:bool=False) -> pl.Path:
        """
          a basic path normalization
          expands user, and resolves the location to be absolute
        """
        result = path
        if symlinks:
            logging.warning("TODO normalize path with symlinks")
        match result.parts:
            case ["~", *xs]:
                result = result.expanduser().resolve()
            case ["/", *xs]:
                result = result
            case _ if root:
                result = (root / path).expanduser().resolve()
            case _:
                pass

        return result

class Walker_m:
    """ A Mixin for walking directories,
      written for py<3.12
      """
    control_e = LoopControl

    def walk_all(self, roots, exts, rec=False, fn=None) -> Generator[dict]:
        """
        walk all available targets,
        and generate unique names for them
        """
        result = []
        match rec:
            case True:
                for root in roots:
                    result += self.walk_target_deep(root, exts, fn)
            case False:
                for root in roots:
                    result += self.walk_target_shallow(root, exts, fn)

        return result

    def walk_target_deep(self, target, exts, fn) -> Generator[pl.Path]:
        logging.info("Deep Walking Target: %s : exts=%s", target, exts)
        if not target.exists():
            return None

        queue = [target]
        while bool(queue):
            current = queue.pop()
            if not current.exists():
                continue
            if current.name in walk_ignores:
                continue
            if current.is_dir() and any([(current / x).exists() for x in walk_halts]):
                continue
            if bool(exts) and current.is_file() and current.suffix not in exts:
                continue
            match fn(current):
                case self.control_e.yes:
                    yield current
                case True if current.is_dir():
                    queue += sorted(current.iterdir())
                case True | self.control_e.yesAnd:
                    yield current
                    if current.is_dir():
                        queue += sorted(current.iterdir())
                case False | self.control_e.noBut if current.is_dir():
                    queue += sorted(current.iterdir())
                case None | False:
                    continue
                case self.control_e.no | self.control_e.noBut:
                    continue
                case _ as x:
                    raise TypeError("Unexpected filter value", x)

    def walk_target_shallow(self, target, exts, fn):
        logging.debug("Shallow Walking Target: %s", target)
        if target.is_file():
            fn_fail = fn(target) in [None, False, self.control_e.no, self.control_e.noBut]
            ignore  = target.name in walk_ignores
            bad_ext = (bool(exts) and (x.is_file() and x.suffix in exts))
            if not (fn_fail or ignore or bad_ext):
                yield target
            return None

        for x in target.iterdir():
            fn_fail = fn(x) in [None, False, self.control_e.no, self.control_e.noBut]
            ignore  = x.name in walk_ignores
            bad_ext = bool(exts) and x.is_file() and x.suffix not in exts
            if not (fn_fail or ignore or bad_ext):
                yield x



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep

def get_random_directions(n=1, random=None):
    """ Choose a direction of cardinal and intercardinal directions """
    dirs = [-1, 0, 1]
    if random is None:
        random = lambda a, x: np.random.choice(a, size=x*2, replace=True, p=None)
    result = random(dirs, n).reshape((n, 2))
    return result

def get_directions(xys):
    """ Given a sequence of points, get the unit direction
    from each point to the next point
    """
    assert(isinstance(xys, np.ndarray))
    assert(len(xys.shape) == 2)
    #convert to vectors:
    #xysPrime.shape = (n, 4)
    #Leading point first to prevent wrap deformation
    # xys_prime = np.column_stack((xys[1:, :], xys[:-1, :]))
    # dx = xys_prime[:, 2] - xys_prime[:, 0]
    # dy = xys_prime[:, 3] - xys_prime[:, 1]
    ds = xys[:-1] - xys[1:]
    #radians:
    arc = np.arctan2(ds[:,1], ds[:,0])
    directions = np.column_stack([np.cos(arc), np.sin(arc)])
    #hypotenuse
    hypos = np.sqrt(np.square(ds[:,0])+np.square(ds[:,1]))
    return np.column_stack((directions, hypos))

def get_unit_vector(p1, p2):
    """ Given two points, get the normalized direction """
    assert(isinstance(p1, np.ndarray))
    assert(isinstance(p2, np.ndarray))
    d = get_distance(p1, p2)
    if np.allclose(d, 0):
        return np.array([0, 0])
    n = (p2-p1)
    normalized = n / d
    return normalized


def is_clockwise(*args, cartesian=True):
    """ Test whether a set of points are in clockwise order  """
    #based on stackoverflow.
    #sum over edges, if positive: CW. negative: CCW
    #assumes normal cartesian of y bottom = 0
    the_sum = 0
    p1s = args
    p2s = list(args[1:])
    p2s.append(args[0])
    pairs = zip(p1s, p2s)
    for p1, p2 in pairs:
        a = (p2[0]-p1[0]) * (p2[1]+p1[1])
        the_sum += a
    if cartesian:
        return the_sum >= 0
    else:
        return the_sum < 0

def is_counter_clockwise(a, b, c):
    """ Given 3 points, do they form a counter clockwise turn """
    assert(all([isinstance(x, np.ndarray) for x in [a, b, c]]))
    offset_b = b - a
    offset_c = c - a
    crossed = np.cross(offset_b, offset_c)
    return crossed >= 0


def __rotate_point_obsolete(p, cen, rads):
    """ Does what rotate point does, explicitly instead
    of with matrix multiplication """
    assert(len(p.shape) == 2)
    c = np.cos(rads)
    s = np.sin(rads)
    centred = p - cen
    cos_p = centred * c
    sin_p = centred * s
    nx = cos_p[:, 0] - sin_p[:, 1]
    ny = sin_p[:, 0] + cos_p[:, 1]
    un_centered = np.column_stack((nx, ny)) + cen
    return un_centered

def rotate_point(p, cen=None, rads=None, rad_min=-QUARTERPI, rad_max=QUARTERPI, random=None):
    """ Given a point, rotate it around a centre point by either radians,
    or within a range of radians
    """
    #p1 = cen, p2=point, @ is matrix mul
    if cen is None:
        cen = np.array([0, 0])
    if rads is None:
        use_radians = random_radian(min_v=rad_min, max_v=rad_max, random=random)
        if isinstance(use_rads, np.ndarray):
            use_radians = use_rads[0]
    else:
        use_radians = rads
    #apply to 1d slices, this allows multiple points to be
    #passed into the function together,
    #without messing up the rotation matmul
    rot_m = rotation_matrix(use_radians)
    offset = (p - cen)
    if len(p.shape) == 1:
        applied = rot_m @ offset
    else:
        applied = np.apply_along_axis(construct_matrix_multiplier(rot_m), 1, offset)
    result = cen + applied
    return result



"""
PatternIterator automates pattern output
"""
##-- imports
from __future__ import annotations
from fractions import Fraction as f
from .utils import time_str
import logging as root_logger

##-- end imports
logging = root_logger.getLogger(__name__)

class PatternIterator:
    """ Automates retrieval of pattern values  """

    def __init__(self, pattern, just_values=True, rand_s=None):
        self.pattern = pattern
        self.denominator = self.pattern.denominator()
        self.position = 0
        self.just_values = just_values
        self.rand_state = rand_s

    def __iter__(self):
        return self

    def __next__(self):
        time = f(self.position, self.denominator)
        self.position += 1
        #if at a new cycle, tick the random generator
        return self.pattern(time, just_values=self.just_values, rand_s=self.rand_state)

    #raise StopIteration



#!/usr/bin/env python
##-- imports
from __future__ import annotations

import logging as root_logger
import pathlib as pl
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)
from uuid import uuid4
##-- end imports

logging = root_logger.getLogger(__name__)

import sh

def convert_alternative(source, output_dir, title):
    target = output_dir / f".{title}.txt"
    logging.info("Converting %s from %s", target, source)
    sh.mutool('mutool', 'convert', '-F', 'text', '-o', str(target), str(source))

def merge_pdfs(paths, output="./pdf_summary"):
    output = pl.Path(output).expanduser().resolve().with_suffix(".pdf")
    writer = PdfWriter()

    if output.exists():
        logging.info("Overwriting existing summary")

    for path in paths:
        try:
            logging.info("File : %s", path.name)
            pdf_obj = PdfReader(path)
            for x in pdf_obj.pages:
                writer.addpage(x)
        except:
            logging.warning("Error Encountered with %s", path)

    writer.write(str(output))

def handle_epub(path, writer, func, temp) -> bool:
    result = False
    if path.suffix == ".epub":
        logging.info("Handling epub: %s", path)
        temp_file_name = pl.Path(temp) / f"{uuid4().hex}.pdf"
        try:
            pandoc.convert_file([str(path)], "pdf", outputfile=str(temp_file_name), format="epub")
            pdf_obj = PdfReader(temp_file_name)
            writer.addpage(func(pdf_obj))
            result = True
        except Exception as err:
            logging.info("Epub Handling Failed: %s", path)
            logging.debug("Epub Error: %s", err)

    return result

def uncompress_pdf(path, writer, func, temp):
    if path.suffix != ".pdf":
        return False

    logging.warning("Attempting to uncompress then add: %s", path)
    temp_unc   = pl.Path(temp) / f"{path.stem}_uncompressed.pdf"
    temp_comp  = pl.Path(temp) / f"{path.stem}_compressed.pdf"
    result = run(['pdftk', str(path), "output", str(temp_unc), "uncompress"],
                 capture_output=True)
    # result = run(['pdftk', str(temp_unc), "output", str(temp_comp), "uncompress"],
    #              capture_output=True)
    return add_to_writer(temp_unc,writer, func, temp)

def add_simple_text_to_writer(path, writer, func, temp):
    logging.warning("Adding Simple Text to Writer: %s", path)
    # from stackoverflow.com/questions/2365411
    # if not path.isascii():
    #     path = unicodedata.normalize("NFKD", path).encode("ascii", "ignore")

    temp_file_name = pl.Path(temp) / f"{uuid4().hex}.pdf"
    simple_path = "/".join(path.parts[-4:])
    pandoc.convert_text(f"File: {simple_path}", "pdf", outputfile=str(temp_file_name), format="md")
    pdf_obj = PdfReader(temp_file_name)
    writer.addpage(func(pdf_obj))
    return True

def finalise_writer(output, writer, base_name, count, bound, force=False) -> tuple[int, PdfWriter]:
    if force or len(writer.pagearray) > bound:
        logging.warning("Writing and incrementing")
        writer.trailer.Info = IndirectPdfDict(
            Title=f'Pdf Summary of {base_name} : Number: {count}',
            Author='JG',
            Subject='pdf summary',
            Creator='doot.pdf.manipluate.summarise_to_pdfs',
        )
        if output.with_stem(f"{output.stem}_{count}").exists():
            logging.info("Overwriting existing summary")
        writer.write(output.with_stem(f"{output.stem}_{count}"))
        writer = PdfWriter()
        count += 1

    return count, writer


def file_is_finished(path) -> bool:
    result = False
    if not path.exists():
        return result

    response = sh.tail('tail', '-n', '1', str(path), capture_output=True, shell=False)
    line = response.stdout.decode()
    if line == END_LINE:
        result = True

    return result

def pdftk_md(path) -> str:
    result = ""
    try:
        response = sh.pdftk(str(path), "dump_data_utf8", capture_output=True, shell=False)
        result   = response.stdout.decode() if response.returncode == 0 else response.stderr.decode()
        result   = "\n".join([x for x in result.split("\n") if "Info" in x])
    except Exception as err:
        result = str(err)

    return result



""" The Top Level DCEL DataStructure. """
##-- imports
from __future__ import annotations
import logging as root_logger
import pickle
from collections import namedtuple
from dataclasses import InitVar, dataclass, field
from itertools import cycle, islice
from math import atan2, degrees
from os.path import isfile
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import numpy as np
import pyqtree

from .constants import EdgeE, FaceE, VertE
from .face import Face
from .halfedge import HalfEdge
from .line_intersector import LineIntersector
from .vertex import Vertex

##-- end imports

logging = root_logger.getLogger(__name__)

#for importing data into the dcel:
DataPair = namedtuple('DataPair', 'key obj data')

@dataclass
class DCELState:
    """ The total DCEL data structure, stores vertices, edges, and faces,
    Based on de Berg's Computational Geometry
    """
    bbox                : np.ndarray     = field(default=np.array([-200, -200, 200, 200]))
    vertices            : Set[Vertex]    = field(default_factory=set)
    faces               : Set[Face]      = field(default_factory=set)
    half_edges          : Set[HalfEdge]  = field(default_factory=set)

    #todo               : make this a stack of quadtrees

    quad_tree_stack     : List[Any]      = field(default_factory=list)
    frontier            : Set[Any]       = field(default_factory=set)
    should_merge_stacks : bool           = field(default=True)
    data                : Dict[Any, Any] = field(default_factory=dict)

    vertex_quad_tree    : pyqtree        = field(init=False, default=None)
    Vertex                               = Vertex
    HalfEdge                             = HalfEdge
    Face                                 = Face

    def __post_init__(self, bbox=None):
        assert(isinstance(bbox, np.ndarray))
        assert(len(bbox) == 4)
        #todo: make this a stack of quadtrees
        self.vertex_quad_tree = pyqtree.Index(bbox=self.bbox)


    def __enter__(self):
        """ Makes the Dcel a reusable context manager, that pushes
        and pops vertex quad trees for collision detection """
        self.push_quad_tree()

    def __exit__(self, e_type, value, traceback):
        self.pop_quad_tree()

    def __str__(self):
        """ Create a text description of the DCEL """
        #pylint: disable=too-many-locals
        vertices_description      = "Vertices: num: {}".format(len(self.vertices))
        edges_description         = "HalfEdges: num: {}".format(len(self.half_edges))
        faces_description         = "Faces: num: {}".format(len(self.faces))

        all_vertices              = [x.getVertices() for x in self.half_edges]
        flattened_vertices        = [x for (x, y) in all_vertices for x in (x, y)]
        set_of_vertices           = set(flattened_vertices)
        vertex_set                = "Vertex Set: num: {}/{}/{}".format(len(set_of_vertices),
                                                        len(flattened_vertices),
                                                        len(self.vertices))

        infinite_edges            = [x for x in self.half_edges if x.is_infinite()]
        infinite_edge_description = "Infinite Edges: num: {}".format(len(infinite_edges))

        complete_edges            = set()
        for x in self.half_edges:
            if not x in complete_edges and x.twin not in complete_edges:
                complete_edges.add(x)

        complete_edge_description        = "Complete Edges: num: {}".format(len(complete_edges))

        edgeless_vertices                = [x for x in self.vertices if x.is_edgeless()]
        edgeless_vertices_description    = "Edgeless vertices: num: {}".format(len(edgeless_vertices))

        edge_count_for_faces             = [str(len(f.edge_list)) for f in self.faces]
        edge_count_for_faces_description = \
                "Edge Counts for Faces: {}".format("-".join(edge_count_for_faces))

        purge_verts = "Verts to Purge: {}".format(len([x for x in self.vertices
                                                       if x.marked_for_cleanup]))
        purge_edges = "Hedges to Purge: {}".format(len([x for x in self.half_edges
                                                        if x.marked_for_cleanup]))
        purge_faces = "Faces to Purge: {}".format(len([x for x in self.faces
                                                       if x.marked_for_cleanup]))

        return "\n".join(["---- DCEL Description: ",
                          vertices_description,
                          edges_description,
                          faces_description,
                          vertex_set,
                          infinite_edge_description,
                          complete_edge_description,
                          edgeless_vertices_description,
                          edge_count_for_faces_description,
                          "-- Purging:",
                          purge_verts,
                          purge_edges,
                          purge_faces,
                          "----\n"])


    #------------------------------
    # def IO
    #------------------------------


    def __repr__(self):
        return "<DECL>"





#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import networkx as nx

@dataclass
class ComponentWriter:
    """
    Simple interface to buffer writing tweets to a json file
    """
    dir_s        : pl.Path   = field()
    name_stem    : str       = field(default="component_{}")
    id_s         : str       = field(default_factory=lambda: str(uuid1()))
    suffix       : str       = field(default=".json")

    stored       : list[Any] = field(default_factory=list)
    tweet_ids    : set[str]  = field(default_factory=set)
    user_ids     : set[str]  = field(default_factory=set)
    write_count  : int       = field(default=20)
    state        : str       = field(default="pre")

    write_states : ClassVar[list[str]] = ["pre", "writing_tweets", "mid", "writing_users", "finished"]

    def __hash__(self):
        return id(self)

    @property
    def path(self):
        return (self.dir_s / self.name_stem.format(self.id_s)).with_suffix(self.suffix)

    def finish(self):
        """ Add the final ] to the file """
        self._maybe_dump(force=True)
        with open(self.path, 'a') as f:
            f.write("\n    ]\n}")
        self.state = "finished"

    def add(self, data, data_type="tweets"):
        """ Add a tweet lazily into the component file """
        match self.state, data_type:
            case "pre" | "writing_tweets", "tweets":
                self.stored.append(data)
                self.tweet_ids.add(data['id_str'])
                self._maybe_dump()
            case "pre" | "writing_tweets", "users":
                logging.debug("Switching Writer to users")
                self._maybe_dump(force=True)
                self.state = "mid"
                self.user_ids.add(data['id_str'])
                self.stored.append(data)
                self._maybe_dump()
            case "mid" | "writing_users", "users":
                self.stored.append(data)
                self.user_ids.add(data['id_str'])
                self._maybe_dump()
            case _:
                raise TypeError("Unexpected State for writer", self.state, data_type)

    def _maybe_dump(self, force=False):
        """
        Dump queued tweets into the component file
        """
        assert(not self.state == "finished")
        if (not force) and len(self.stored) < self.write_count:
            return
        if not bool(self.stored):
            return

        # convert to str, chop of brackets
        write_str = json.dumps(self.stored, indent=4)[2:-2]
        with open(self.path, 'a') as f:
            match self.state:
                case "pre":
                    f.write("{\n    \"tweets\": [\n")
                    self.state = "writing_tweets"
                case "writing_tweets" | "writing_users":
                    f.write(",\n")
                case "mid":
                    f.write("\n    ],\n    \"users\": [\n")
                    self.state = "writing_users"

            f.write(write_str)

        self.stored.clear()

    def summary(self):
        tweet_ids = " ".join(self.tweet_ids)
        user_ids  = " ".join(self.user_ids)
        comp_name = self.path.stem
        return f"Component: {comp_name} Counts: [{len(self.tweet_ids)} {len(self.user_ids)}] TweetIds: [{tweet_ids}] UserIds: [{user_ids}]"



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)


##-- end imports

logging = logmod.getLogger(__name__)

@register_class("exit")
class ExitCmd:
    """
    Exit the repl, automatically saving the self state
    """

    def __call__(self, line):
        # TODO add option for saving location
        logging.info("Quitting")
        filename = "repl.auto"
        self._repl.state.engine.save_file(filename)
        return True



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv.decorators.base import JGDVBaseDecorator

class KWrapper(JGDVBaseDecorator):
    """ Decorators for actions
    Kwrapper is accessible as DootKey.kwrap

    It registers arguments on an action and extracts them from the spec and state automatically.

    provides: expands/paths/types/requires/returns/args/kwargs/redirects/redirects_many

    The kwarg 'hint' takes a dict and passes the contents to the relevant expansion method as kwargs

    arguments are added to the tail of the action args, in order of the decorators.
    the name of the expansion is expected to be the name of the action parameter,
    with a "_" prepended if the name would conflict with a keyword., or with "_ex" as a suffix
    eg: @DootKey.kwrap.paths("from") -> def __call__(self, spec, state, _from):...
    or: @DootKey.kwrap.paths("from") -> def __call__(self, spec, state, from_ex):...
    """

    @staticmethod
    def _annotate_keys(f, keys:list) -> bool:
        """ cache original args, and cache declared keys """
        if hasattr(f, FUNC_WRAPPED): # Deal with the actual function, not any decorators
            return KWrapper._annotate_keys(f.__wrapped__, keys)

        if not hasattr(f, ORIG_ARGS): # store the original arguments for easy access
            setattr(f, ORIG_ARGS, f.__code__.co_varnames[:f.__code__.co_argcount])

        if not hasattr(f, KEY_ANNOTS): # ensure theres a place for annotations
            setattr(f, KEY_ANNOTS, [])

        # prepend annotations, so written decorator order is the same as written arg order:
        # (ie: @wrap(x) @wrap(y) @wrap(z) def f (x, y, z), even though z's decorator is applied first
        new_annotations = keys + getattr(f, KEY_ANNOTS)
        setattr(f, KEY_ANNOTS, new_annotations)

        # run the key check
        if not KWrapper._check_keys(f, getattr(f, KEY_ANNOTS)):
            raise doot.errors.DootKeyError("Annotations do not match signature", getattr(f, ORIG_ARGS, []), getattr(f, KEY_ANNOTS), f.__qualname__)

        return True

    @staticmethod
    def _annotate_non_expansions(f, keys:list, type_="in") -> bool:
        """
        Annotate required inputs and output
        """
        if hasattr(f, FUNC_WRAPPED):
            return KWrapper._annotate_non_expansions(f.__wrapped__, keys)

    @staticmethod
    def _check_keys(f, keys, offset=0) -> bool:
        """ test declared args to a list of keys """
        if hasattr(f, ORIG_ARGS):
            code_args           = getattr(f, ORIG_ARGS)
            code_argcount       = len(code_args)
        else:
            code_argcount           = f.__code__.co_argcount
            code_args               = f.__code__.co_varnames[:code_argcount]

        result                  = True
        if code_args[0]         == "self":
            code_args           = code_args[1:]

        # First two params should always be spec and state
        result &= code_args[:2] == ("spec", "state")

        # The rest should match keys
        for actual, expected in zip(code_args[:1+offset:-1], keys[::-1]):
            match expected:
                case DootMultiKey():
                    pass
                case DootSimpleKey() | str() if actual.startswith("_"):
                    pass
                case DootSimpleKey() | str():
                    result &= ((actual == expected) or (actual == f"{expected}_ex"))

        return result

    @staticmethod
    def _add_key_handler(f):
        """ idempotent key handler so decorated functions dont add unnecessary stack frames """
        if DootDecorator.has_annotations(f):
            return f

        DootDecorator.truncate_signature(f)

        match getattr(f, ORIG_ARGS)[0]:
            case "self": # The method form handler

                @ftz.wraps(f)
                def action_expands(self, spec, state, *call_args, **kwargs):
                    try:
                        expansions = [x(spec, state) for x in getattr(f, KEY_ANNOTS)]
                    except KeyError as err:
                        printer.warning("Action State Expansion Failure: %s", err)
                        return False
                    all_args = (*call_args, *expansions)
                    return f(self, spec, state, *all_args, **kwargs)
            case _: # The function form handler

                @ftz.wraps(f)
                def action_expands(spec, state, *call_args, **kwargs):
                    try:
                        expansions = [x(spec, state) for x in getattr(f, KEY_ANNOTS)]
                    except KeyError as err:
                        printer.warning("Action State Expansion Failure: %s", err)
                        return False
                    all_args = (*call_args, *expansions)
                    return f(spec, state, *all_args, **kwargs)

        DootDecorator.annotate(action_expands, {KEYS_HANDLED})
        return action_expands

    @staticmethod
    def taskname(f):
        KWrapper._annotate_keys(f, [DootKey.build(STATE_TASK_NAME_K, exp_hint="type")])
        return KWrapper._add_key_handler(f)

    @staticmethod
    def expands(*args, hint:dict|None=None, **kwargs):
        """ mark an action as using expanded string keys """
        exp_hint = {"expansion": "str", "kwargs" : hint or {} }
        keys = [DootKey.build(x, exp_hint=exp_hint, **kwargs) for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_keys(f, keys)
            return KWrapper._add_key_handler(f)

        return expand_wrapper

    @staticmethod
    def paths(*args, hint:dict|None=None, **kwargs):
        """ mark an action as using expanded path keys """
        exp_hint = {"expansion": "path", "kwargs" : hint or {} }
        keys = [DootKey.build(x, exp_hint=exp_hint, **kwargs) for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_keys(f, keys)
            return KWrapper._add_key_handler(f)

        return expand_wrapper

    @staticmethod
    def types(*args, hint:dict|None=None, **kwargs):
        """ mark an action as using raw type keys """
        exp_hint = {"expansion": "type", "kwargs" : hint or {} }
        keys = [DootKey.build(x, exp_hint=exp_hint, **kwargs) for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_keys(f, keys)
            return KWrapper._add_key_handler(f)

        return expand_wrapper

    @staticmethod
    def args(f):
        """ mark an action as using spec.args """
        # TODO handle expansion hint for the args
        KWrapper._annotate_keys(f, [DootArgsKey("args")])
        return KWrapper._add_key_handler(f)

    @staticmethod
    def kwargs(f):
        """ mark an action as using spec.args """
        KWrapper._annotate_keys(f, [DootKwargsKey("kwargs")])
        return KWrapper._add_key_handler(f)

    @staticmethod
    def redirects(*args):
        """ mark an action as using redirection keys """
        keys = [DootKey.build(x, exp_hint="redirect") for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_keys(f, keys)
            return KWrapper._add_key_handler(f)

        return expand_wrapper

    @staticmethod
    def redirects_many(*args, **kwargs):
        """ mark an action as using redirection key lists """
        keys = [DootKey.build(x, exp_hint="redirect_multi") for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_keys(f, keys)
            return KWrapper._add_key_handler(f)

        return expand_wrapper

    @staticmethod
    def requires(*args, **kwargs):
        """ mark an action as requiring certain keys to be passed in """
        keys = [DootKey.build(x, **kwargs) for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_non_expansions(f, keys)
            return f

        return expand_wrapper

    @staticmethod
    def returns(*args, **kwargs):
        """ mark an action as needing to return certain keys """
        keys = [DootKey.build(x, **kwargs) for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_non_expansions(f, keys)
            return f

        return expand_wrapper

    @staticmethod
    def references(*args, **kwargs):
        """ mark keys to use as to_coderef imports """
        exp_hint = {"expansion": "coderef", "kwargs" : {} }
        keys = [DootKey.build(x, exp_hint=exp_hint, **kwargs) for x in args]

        def expand_wrapper(f):
            KWrapper._annotate_keys(f, keys)
            return KWrapper._add_key_handler(f)

        return expand_wrapper



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import datetime
import scrapy
from scrapy import signals
from itemadapter import is_item, ItemAdapter
from scrapy.utils.httpobj import urlparse_cached
import gzip
import pickle
from time import time
from w3lib.http import headers_dict_to_raw, headers_raw_to_dict
from bs4 import BeautifulSoup
from urllib.parse import urlparse

import scrapy
from scrapy.http import Headers, Response
from scrapy.http.request import Request
from scrapy.responsetypes import responsetypes
from scrapy.spiders import Spider
from scrapy.utils.httpobj import urlparse_cached
from itemadapter import ItemAdapter
from scrapy.exporters import XmlItemExporter
from scrapy.crawler import CrawlerProcess

# available in scrapy.exporters:
# XmlItemExporter(file, item_element='item', root_element='items', **kwargs)
# CsvItemExporter(file, include_headers_line=True, join_multivalued=',', errors=None, **kwargs)
# PickleItemExporter(file, protocol=0, **kwargs)
# PprintItemExporter(file, **kwargs)
# JsonItemExporter(file, **kwargs)
# JsonLinesItemExporter(file, **kwargs)

class ItemData(scrapy.Item):
    # define the fields for your item here like:
    # name = scrapy.Field()
    pass

class SimpleXMLExporter:
    """
    """

    def open_spider(self, spider):
        self.items = []

    def close_spider(self, spider):
        with open(spider.locs.build / "results.xml", "wb") as f:
            exporter = XmlItemExporter(f)
            exporter.start_exporting()
            for item in self.items:
                exporter.export_item(item)
            exporter.finish_exporting()

    def process_item(self, item, spider):
        self.items.append(item)
        return item

class RawXMLExporter:
    """
    write raw xml into a file
    """

    def open_spider(self, spider):
        self.base_loc = spider.locs.data
        self.soups = {}

    def close_spider(self, spider):
        # write the soup
        for name, soup in self.soups.items():
            soup_file = self.soup_file(name)
            soup_file.write_text(soup.prettify())

    def process_item(self, item, spider):
        if not all(x in item for x in ["source_url", "data"]):
            logging.info("Not XML exporting item: %s", item)
            return item

        soup = self.maybe_create_soup(item['source_url'])

        # add to the soup
        match item['data']:
            case str() as text:
                xml_data                   = BeautifulSoup(text, features="lxml-xml")
            case list() as the_list:
                xml_data                   = BeautifulSoup("\n".join(the_list), features="lxml-xml")
            case _ as unknown:
                logging.warning("Unknown form to save as xml: %s", type(unkown))
                xml_data                   = BeautifulSoup(str(unkown), features="lxml-xml")

        xml_item                   = xml_data.wrap(soup.new_tag("item"))
        xml_item['source_url']     = item['source_url']
        xml_item['parse_date']     = datetime.datetime.now().isoformat()
        xml_item['needs_subsplit'] = item.get('needs_subsplit', False)

        soup.items.append(xml_item)
        return item

    def soup_file(self, netloc:str):
        return (self.base_locs / netloc.replace(".","_")).with_suffix(".xml")

    def maybe_create_soup(self, url:str):
        base_url = urlparse(url).netloc
        if base_url in self.soups:
            return self.soups[base_url]
        soup_file = self.soup_file(base_url)

        if soup_file.exists():
            soup = BeautifulSoup(soup_file.read_text(), features="lxml-xml")
        else:
            soup = BeautifulSoup(features="lxml-xml")
            soup.append(soup.new_tag("items"))
            soup.items['creation_date'] = datetime.datetime.now().isoformat()

        self.soups[base_url] = soup
        return soup

class ItemPipeline:

    # def open_spider(self, spider)
    # def close_spider(self, spider)
    # def from_crawler(cls, crawler)

    def process_item(self, item, spider):
        # must return item, deferred, or raise scrapy.exceptions.DropItem
        return item



#!/usr/bin/env python3
"""
Construct.py based readers for bioware infinity engine file formats:
CHITIN.KEY
(todo) .BIF
(todo) .DLG
(todo) .TLK
(todo) .2DA
(todo) .BS
(todo) .BCS
(todo) .MAZE
(todo) .STR

https://gibberlings3.github.io/iesdp/file_formats/index.htm

Notes:
numbers are little endian
"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import construct as C

##-- file types
infinity_file_types = {
    0x0001 : ".bmp",
    0x0002 : ".mve",
    0x0004 : ".wav",
    0x0005 : ".wfx",
    0x0006 : ".plt",
    0x03e8 : ".bam",
    0x03e9 : ".wed",
    0x03ea : ".chu",
    0x03eb : ".tis",
    0x03ec : ".mos",
    0x03ed : ".itm",
    0x03ee : ".spl",
    0x03ef : ".bcs",
    0x03f0 : ".ids",
    0x03f1 : ".cre",
    0x03f2 : ".are",
    0x03f3 : ".dlg",
    0x03f4 : ".2da",
    0x03f5 : ".gam",
    0x03f6 : ".sto",
    0x03f7 : ".wmp",
    0x03f8 : ".chr",
    0x03f9 : ".bs",
    0x03fa : ".chr",
    0x03fb : ".vvc",
    0x03fc : ".vef",
    0x03fd : ".pro",
    0x03fe : ".bio",
    0x03ff : ".wbm",
    0x0400 : ".fnt",
    0x0402 : ".gui",
    0x0403 : ".sql",
    0x0404 : ".pvrz",
    0x0405 : ".glsl",
    0x0408 : ".menu",
    0x0409 : ".menu",
    0x040a : ".ttf",
    0x040b : ".png",
    0x040c : ".bah",
    0x0802 : ".ini",
    0x0803 : ".src",
    }

##-- end file types

class BioWareInfinityBinaryMixin:
    """
    Mixin for building parts of infinity engine file formats

    """
    Word   = C.Int16ul
    DWord  = C.Int32ul
    file_types = infinity_file_types

    ##-- top level
    def build_key_v1_format(self) -> C.Struct:
        """
        Build a construct parser for infinity engine KEY format file
        Structure:
        header, [bif descriptions], [resource descriptions]

        see:
        https://gibberlings3.github.io/iesdp/file_formats/ie_formats/key_v1.htm
        """
        return C.Struct("header" / self._key_v1_header,
                        C.Seek(C.this.header.bif_offset),
                        "bif_descs" / self._key_v1_bif_description[C.this.header.bif_num],
                        C.Seek(C.this.header.res_offset),
                        "res_descs" / self._key_v1_resource_description[C.this.header.res_num]
                        )

    def build_bif_v1_format(self) -> C.Struct:
        """
        https://gibberlings3.github.io/iesdp/file_formats/ie_formats/bif_v1.htm
        """
        return C.Struct("header" / self._bif_v1_header,
                        C.Seek(C.this.header.offset),
                        "file_entries" / self._bif_v1_entry[C.this.header.file_entries_count],
                        C.Check(C.len_(C.this.file_entries) == C.this.header.file_entries_count),
                        "tileset_entries" / self._bif_v1_tileset[C.this.header.tileset_entries_count],
                        C.Check(C.len_(C.this.tileset_entries) == C.this.header.tileset_entries_count),
                        )

    def build_bifc_v1_format(self) -> C.Struct:
        return C.Struct("header" / self._bif_compressed_v1_header,
                        "blocks" / C.GreedyRange(self._bif_compressed_v1_block),
                        )

    def build_tlk_v1_format(self) -> C.Struct:
        """
        https://gibberlings3.github.io/iesdp/file_formats/ie_formats/tlk_v1.htm
        """
        return C.Struct("header" / self._tlk_v1_header,
                        C.Seek(18),
                        "entries" / self._tlk_v1_entries[C.this.header.entry_count],
                        )

    def build_dlg_v1_format(self) -> C.Struct:
        """
        https://gibberlings3.github.io/iesdp/file_formats/ie_formats/dlg_v1.htm
        """
        return C.Struct("header"              / self._dlg_v1_header,
                        "state_table"         / C.Pointer(C.this.header.state_offset,              self._dlg_v1_state[C.this.header.state_count]),
                        "transition_table"    / C.Pointer(C.this.header.transition_offset,         self._dlg_v1_transition[C.this.header.transition_count]),
                        "state_triggers"      / C.Pointer(C.this.header.state_trigger_offset,      self._dlg_v1_data[C.this.header.state_trigger_count]),
                        "transition_triggers" / C.Pointer(C.this.header.transition_trigger_offset, self._dlg_v1_data[C.this.header.transition_trigger_count]),
                        "actions"             / C.Pointer(C.this.header.action_offset,             self._dlg_v1_data[C.this.header.action_count])
                        )

    def build_2da_format(self) -> C.Struct:
        """
        https://gibberlings3.github.io/iesdp/file_formats/ie_formats/2da.htm
        TODO detect 0xff for encryption
        """
        raise NotImplementedError()

    def build_src_format(self) -> C.Struct:
        """
        https://gibberlings3.github.io/iesdp/file_formats/ie_formats/src.htm
        """
        return C.Struct("entry_count" / self.DWord,
                        "entries" / C.Struct("str_ref" / self.DWord,
                                             "weight"  / self.DWord)[C.this.entry_count]
                        )

    ##-- end top level

    ##-- components

    @property
    def _key_v1_header(self) -> C.Struct:
        return C.Struct(C.Const(b"KEY"), C.Padding(1),
                        "version"    / C.FixedSized(4, C.NullStripped(C.GreedyString("ascii"))),
                        "bif_num"    / self.DWord,
                        "res_num"    / self.DWord,
                        "bif_offset" / self.DWord,
                        "res_offset" / self.DWord,
                        )

    @property
    def _key_v1_bif_description(self) -> C.Struct:
        return C.Struct("pos" / C.Tell,
                        "id" / C.Computed(C.this._index),
                        "file_len"    / self.DWord,
                        "name_offset" / self.DWord,
                        "name"        / C.Pointer(C.this._.name_offset, C.CString("ascii")),
                        "name_len"    / self.Word,
                        "location"    / C.FlagsEnum(self.Word, data=1, cache=2, cd1=4, cd2=8, cd3=16, cd5=24, cd6=32),
                        )

    @property
    def _key_v1_resource_description(self) -> C.Struct:
        return C.Struct("id"           / C.Computed(C.this._index),
                        "name"         / C.PaddedString(8, "ascii"),
                        "type"         / self.Word,
                        "resource_key" / self._key_v1_resource_key
                        )

    @property
    def _key_v1_resource_key(self) -> C.Struct:
        return C.ByteSwapped(C.BitStruct(
            "bif"  / C.BitsInteger(12),
            "tile" / C.BitsInteger(6),
            "file" / C.BitsInteger(14)
        ))

    @property
    def _bif_v1_header(self) -> C.Struct:
        return C.Struct(C.Const(b"BIFF"), C.Const(b'V1  '),
                        "file_entries_count"    / self.DWord,
                        "tileset_entries_count" / self.DWord,
                        "offset"                / self.DWord, # from start of file
                        )

    @property
    def _bif_v1_entry(self) -> C.Struct:
        return C.Struct("resource_key" / self._key_v1_resource_key,
                        "offset"       / self.DWord, # from start of file
                        "size"         / self.DWord,
                        "type"         / self.Word,
                        self.Word,
                        "data"         / C.Pointer(C.this.offset, C.Array(C.this.size, C.Byte)),
                        )

    @property
    def _bif_v1_tileset(self) -> C.Struct:
        return C.Struct("resource_key" / self._key_v1_resource_key,
                        "offset"       / self.DWord,
                        "count"        / self.DWord,
                        "size"         / self.DWord,
                        "type"         / C.Const(b'\x3eb'),
                        self.Word
                        # Data is ignored for now
                        )

    @property
    def _bif_compressed_v1_header(self) -> C.Struct:
        return C.Struct(C.Const(b"BIFC"), c.Const(b'V1.0'),
                        "size" / self.DWord
                        )

    @property
    def _bif_compressed_v1_block(self) -> C.Struct:
        return C.Struct("decompressed_size" / self.DWord,
                        "compressed_size"   / self.DWord,
                        "data"              / C.FixedSized(C.this.compressed_size, C.Compressed(C.GreedyBytes, "zlib")),
                        )

    @property
    def _tlk_v1_header(self) -> C.Struct:
        # 4 dword : offset from SoF to string data
        return C.Struct(C.Const(b"TLK "), C.Const(b"V1  "),
                        "language"      / self.Word,
                        "entry_count"   / self.DWord,
                        "string_offset" / self.DWord,
                        )

    @property
    def _tlk_v1_entries(self) -> C.Struct:
        metadata = C.FlagsEnum(self.Word,
                               message_data=1,
                               sound=2,
                               standard_message=4,
                               token=8
                               )

        return C.Struct("index"           / C.Computed(C.this._index),
                        "position"        / C.Tell,
                        "metadata"        / metadata,
                        "sound"           / self._tlk_v1_resref,
                        "volume"          / self.DWord,
                        "pitch"           / self.DWord,
                        "relative_offset" / self.DWord,
                        "length"          / self.DWord,
                        "full_offset"     / C.Computed(C.this._.header.string_offset + C.this.relative_offset),
                        # C.Probe(),
                        "data"            / C.Pointer(C.this.full_offset, C.FixedSized(C.this.length, C.NullStripped(C.GreedyString("utf-8")))),
                        # C.Probe(C.this.data),
                        )

    @property
    def _tlk_v1_strref(self) -> C.Struct:
        return C.Array(4, C.Byte)

    @property
    def _tlk_v1_resref(self) -> C.Struct:
        return C.FixedSized(8, C.Byte)

    @property
    def _dlg_v1_header(self) -> C.Struct:
       return C.Struct(C.Const(b"DLG "), C.Const(b"V1.0"),
                       "state_count"               / self.DWord,
                       "state_offset"              / self.DWord,
                       "transition_count"          / self.DWord,
                       "transition_offset"         / self.DWord,
                       "state_trigger_offset"      / self.DWord,
                       "state_trigger_count"       / self.DWord,
                       "transition_trigger_offset" / self.DWord,
                       "transition_trigger_count"  / self.DWord,
                       "action_offset"             / self.DWord,
                       "action_count"              / self.DWord,
                       # "interruption"              / C.Enum(self.DWord, ),
                    )

    @property
    def _dlg_v1_state(self) -> C.Struct:
        return C.Struct("actor_response_string_reference" / self._tlk_v1_strref,
                        "transition_index"  / self.DWord,
                        "transition_count"  / self.DWord, # state transitions = transition_index -> (transition_index + transition_count) - 1
                        "trigger_index"     / self.DWord) # 0xFFFFFFFF if no triggers

    @property
    def _dlg_v1_transition(self) -> C.Struct:
        metadata = C.FlagsEnum(self.DWord,
                               associated_text=1,
                               trigger=2,
                               action=4,
                               terminates=8,
                               journal=16,
                               interrupt=32,
                               add_unsolved_quest_journal=64,
                               add_journal=128,
                               add_solved_quest_journal=256,
                               delayed_script_execution=512,
                               clear_actions=1024,
                               )

        return C.Struct(
            "metadata"                    / metadata,
            "transition_string_reference" / self._tlk_v1_strref,
            "journal_string_reference"    / self._tlk_v1_strref,
            "trigger_index"               / self.DWord,
            "action_index"                / self.DWord,
            "next_dlg_resource_ref"       / self._tlk_v1_resref,
            "next_state_index"            / self.DWord
        )

    @property
    def _dlg_v1_data(self) -> C.Struct:
        # state trigger / transition triggr / action
        return C.Struct("offset" / self.DWord,
                        "length" / self.DWord,
                        "text"   / C.Pointer(C.this.offset, C.FixedSized(C.this.length, C.NullStripped(C.GreedyString("ascii")))),
                        )

    ##-- end components



##-- imports
from __future__ import annotations
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Generic, Iterable,
                    Iterator, Mapping, Match, MutableMapping, Protocol,
                    Sequence, Tuple, TypeAlias, TypeGuard, TypeVar, cast)

from functools import wraps

##-- end imports

from jgdv._interfaces.decorator import JGDVDecorator_i

T = TypeVar('T')

class EnsureDSLInit(JGDVDecorator_i):
    """ Utility Decorator for DSLs  raising error if not initialised """

    def __call__(self, *args, **kwargs):
        if not self._parsers_initialised:
            raise RuntimeError("DSL Not Initialised")

        return self._func(self, *args, **kwargs)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class DCELio:

    def copy(self):
        """ Completely duplicate the dcel """
        new_dcel = DCEL(self.bbox)
        new_dcel.import_data(self.export_data())
        return new_dcel

    def export_data(self):
        """ Export a simple format to define vertices, halfedges, faces,
        uses identifiers instead of objects, allows reconstruction """
        data = {
            'vertices'   : [x._export() for x in self.vertices],
            'half_edges' : [x._export() for x in self.half_edges],
            'faces'      : [x._export() for x in self.faces],
            'bbox'       : self.bbox
        }
        return data

    def import_data(self, data):
        """ Import the data format of identifier links to actual links,
        of export output from a dcel """
        #pylint: disable=too-many-locals
        #pylint: disable=too-many-statements
        assert(all([x in data for x in ['vertices', 'half_edges', 'faces', 'bbox']]))
        self.bbox = data['bbox']

        #dictionarys used to store {newIndex : (newIndex, newObject, oldData)}
        local_vertices = {}
        local_edges = {}
        local_faces = {}
        output_mapping = {}

        #-----
        # Reconstruct Verts, Edges, Faces:
        #-----
        #construct vertices by index
        logging.info("Re-Creating Vertices: {}".format(len(data['vertices'])))
        for v_data in data['vertices']:
            combined_data = {}
            combined_data.update({VertE.__members__[a] : b for a, b in v_data['enum_data'].items()})
            combined_data.update(v_data['non_enum_data'])

            new_vert = Vertex(np.array([v_data['x'], v_data['y']]),
                              index=v_data['i'], data=combined_data,
                              dcel=self, active=v_data['active'])
            logging.debug("Re-created vertex: {}".format(new_vert.index))
            local_vertices[new_vert.index] = DataPair(new_vert.index, new_vert, v_data)

        #edges by index
        logging.info("Re-Creating HalfEdges: {}".format(len(data['half_edges'])))
        for e_data in data['half_edges']:
            combined_data = {}
            combined_data.update({EdgeE.__members__[a] : b for a, b in e_data['enum_data'].items()})
            combined_data.update(e_data['non_enum_data'])
            new_edge = HalfEdge(index=e_data['i'], data=combined_data, dcel=self)
            logging.debug("Re-created Edge: {}".format(new_edge.index))
            local_edges[new_edge.index] = DataPair(new_edge.index, new_edge, e_data)

        #faces by index
        logging.info("Re-Creating Faces: {}".format(len(data['faces'])))
        for f_data in data['faces']:
            combined_data = {}
            combined_data.update({FaceE.__members__[a] : b for a, b in f_data['enum_data'].items()})
            combined_data.update(f_data['non_enum_data'])
            new_face = Face(site=np.array([f_data['sitex'], f_data['sitey']]), index=f_data['i'],
                            data=combined_data, dcel=self)
            logging.debug("Re-created face: {}".format(new_face.index))
            local_faces[new_face.index] = DataPair(new_face.index, new_face, f_data)

        #-----
        # def Upon reconstruction, reattach ids to the same objects
        #-----
        #this only update standard connections, not user connections
        #TODO: PASS OUT A MAPPING OF OLD IDS TO NEW FOR USER UPDATES
        try:
            #connect vertices to their edges
            for vertex in local_vertices.values():
                vertex.obj.half_edges.update( \
                    [local_edges[x].obj for x in vertex.data['half_edges']])
        except Exception:
            logging.warning("Import Error for vertex")

        try:
            #connect edges to their vertices, and neighbours, and face
            for edge in local_edges.values():
                if edge.data['origin'] is not None:
                    edge.obj.origin = local_vertices[edge.data['origin']].obj
                if edge.data['twin'] is not None:
                    edge.obj.twin = local_edges[edge.data['twin']].obj
                if edge.data['next'] is not None:
                    edge.obj.next = local_edges[edge.data['next']].obj
                if edge.data['prev'] is not None:
                    edge.obj.prev = local_edges[edge.data['prev']].obj
                if edge.data['face'] is not None:
                    edge.obj.face = local_faces[edge.data['face']].obj
        except Exception:
            logging.warning("Import Error for edge")

        try:
            #connect faces to their edges
            for face in local_faces.values():
                face.obj.edge_list = [local_edges[x].obj for x in face.data['edges']]
        except Exception:
            logging.warning("Import Error for face")

        #Now recalculate the quad tree as necessary
        self.calculate_quad_tree()

        #todo: pass the mapping back
        output_mapping['verts'] = {x.data['i'] : x.key for x in local_vertices.values()}
        output_mapping['edges'] = {x.data['i'] : x.key for x in local_edges.values()}
        output_mapping['faces'] = {x.data['i'] : x.key for x in local_faces.values()}
        return output_mapping

    @staticmethod
    def loadfile(filename):
        """ Create a DCEL from a saved pickle """
        if not isfile("{}.dcel".format(filename)):
            raise Exception("Non-existing filename to load into dcel")
        with open("{}.dcel".format(filename), 'rb') as f:
            dcel_data = pickle.load(f)
        the_dcel = DCEL()
        the_dcel.import_data(dcel_data)
        return the_dcel

    def savefile(self, filename):
        """ Save dcel data to a pickle """
        the_data = self.export_data()
        with open("{}.dcel".format(filename), 'wb') as f:
            pickle.dump(the_data, f)




#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import abc
import logging as logmod
import traceback
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

logging = logmod.getLogger(__name__)
rst = RP.rst

@register_class("forcep")
class ForceParserCmd:
    """ Force Parser:
    Query the bootstrap parser,
    and if supplied text, parse it and try to run it
    """

    def __init__(self):
        self._parser = self._gen_parser()

    def _gen_parser(self):
        query          = pp.Word(pp.alphas + ".")("query") + pp.Suppress(pp.Literal("?"))
        send_to_parser = rst("send")
        force_parser = query + pp.Optional(send_to_parser)
        return force_parser

    def __call__(self, line):
        try:
            # parse the line
            params = self._parser.parse_string(line, parse_all=True)
        except pp.ParseException as err:
            print("Unrecognised Option: ", err.pstr)
            print(self.__doc__)
            return

        try:
            # Get the handler for the specified signal
            parser = self._repl.state.engine._dsl[params.query]
            print(f"Retrieved: {parser}\n")
            if not bool(params.send):
                print("Nothing sent to parser")
                return

            print(f"Trying Parser on: {params.send}")
            # if exists, parse, then call engine on it
            if isinstance(parser, HandlerSpec_i):
                built = parser.build()
            else:
                assert(hasattr(parser.func, "parse_string"))
                built = parser.func

            dfs_activate(built)
            forced_result = built.parse_string(params.send.strip(), parseAll=True)[:]
            dfs_activate(built, remove=True)

            self._repl.state.debug_data = forced_result
            print(f"----- Forced Parse Result: {forced_result}\n")

            if isinstance(forced_result, tuple):
                forced_result = forced_result[1]

            answer = input("Execute result? Y/* ")
            if answer == "Y":
                print("Attempting to execute result:")
                self._repl.state.ctxs = self._repl.state.engine(forced_result,
                                                              ctxset=self._repl.state.ctxs)
            else:
                print("Not Executing result")

        except pp.ParseException as err:
            traceback.print_tb(err.__traceback__)
            logging.warning(f"Parse Failure: {err.markInputline()}")
            logging.warning(err)
        except Exception as err:
            print("\n")
            traceback.print_tb(err.__traceback__)
            print("\n")
            logging.warning(f"Force Failure: {err}")



"""

"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

import pyparsing as pp

logging     = logmod.getLogger(__name__)
config      = acab.config
# TODO import
Debugger    = config.imports.specific.debug
debug_intro = config.module.REPL.debug.intro

the_debugger = Debugger()

# TODO breakpoint a semantic handler by name,
# TODO breakpoint (in)dependent semantic function
# TODO breakpoint a node by sentence path
# TODO breakpoint an operator/action
# TODO breakpoint a variable
#

@register_class("break")
class BreakCmd:
    """
    Control the Acab Debugger.
    Specified in Config [Module.Debug].

    Usage:
    break file line maybe_funcname
    break rule.name?

    break delete num
    break list

    # To break on semantic execution:

    """

    def __init__(self):
        self._parser   = self._gen_parser()

    def _gen_parser(self):
        filename = pp.Word(pp.alphas + "/_") + pp.Literal(".py")
        filename.set_parse_action(lambda s,l,t: "".join(t[:]))

        linenum  = pp.Word(pp.nums)
        linenum.set_parse_action(lambda s,l,t: int(t.line))

        basic_bp     = filename("file") + pp.Suppress(pp.Keyword(":")) + linenum("line")

        semantic_bp  = pp.Keyword("semantic") + pp.Regex(".+?\?")("semantic")

        parser_bp    = pp.Keyword("parser") + rst("parser")

        break_parser = parser_bp | semantic_bp("semantic") | basic_bp("basic") | rst

        return break_parser

    def __call__(self, line):
        if not bool(the_debugger):
            the_debugger.set_running_trace()

        try:
            ctxs = self._parser.parse_string(line, parse_all=True)
        except pp.ParseException as err:
            print("Unrecognized option: ", err.pstr)
            print(self.__doc__)
            return

        # TODO refactor the basic/semantic logic into the debugger
        if "basic" in ctxs:
            self.handle_basic(ctxs)
        elif "semantic" in ctxs:
            self.handle_semantic(ctxs)
        elif "parser" in ctxs:
            self.handle_parser(ctxs)
        else:
            repl   = self._repl
            state  = self._repl.state
            engine = self._repl.state.engine
            ctxs   = self._repl.state.ctxs
            print("")
            print("\n\t".join(debug_intro))
            print("")
            the_debugger.set_trace()

    def handle_basic(self, ctxs):
        bp_result = the_debugger.set_break(ctxs.file, ctxs.line)
        if bp_result is None:
            print(f"Breakpoint Set: {ctxs.file} : {ctxs.line}")
        else:
            print(f"Breakpoint NOT Set: {bp_result}")

    def handle_parser(self, ctxs):
        # TODO add debug breakpoint to a parser
        # Either import and inspect, print all parser elements, if query
        if ctxs['parser'][-1] == "?":
            module = importlib.import_module(ctxs['parser'].strip()[:-1])
            parsers = [x for x in dir(module) if isinstance(getattr(module, x), pp.ParserElement)]

            if not bool(parsers):
                return print("No Available Parsers to Breakpoint on.")

            print(f"Available Parsers in {ctxs['parser'][:-1]}")
            for x in parsers:
                print(f"-- {x}")
            return

        # When specific, flip flop:
        mod_str, p_str = splitext(ctxs['parser'].strip())
        module = importlib.import_module(mod_str)
        if not hasattr(module, p_str[1:]) or not isinstance(getattr(module, p_str[1:]), pp.ParserElement):
            print(f"No parser named {p_str[1:]} in {mod_str}.")
            print("Try quering the  with a trailing '?'")
            return

        parser : pp.ParserElement = getattr(module, p_str[1:])
        turn_off = hasattr(parser._parse, "_originalParseMethod")
        if turn_off:
            print(f"Turning Breakpoint off for: {ctxs['parser']}")
            parser.set_break(False)
        else:
            print(f"Turning Breakpoint on for: {ctxs['parser']}")
            parser.set_break(True)

    def handle_semantic(self, ctxs):
        # run query
        self._repl.state.ctxs = self._repl.state.engine(ctxs.semantic)
        # attach semantic breakpoints to each prod_abstraction
        if (len(self._repl.state.ctxs) == 1 and
            isinstance(self._repl.state.ctxs[0]._current.value, Instruction_i)):
            curr = self._repl.state.ctxs[0]._current.value
            curr.breakpoint = not curr.breakpoint
            if curr.breakpoint:
                print(f"Breakpoint Set: {repr(curr)} : {curr.uuid}")
            else:
                print(f"Breakpoint Unset: {repr(curr)} : {curr.uuid}")

        elif bool(self._repl.state.ctxs):
            count = 0
            for inst in self._repl.state.ctxs:
                for bind in inst.data.items():
                    if isinstance(bind, Instruction_i):
                        bind.breakpoint = not bind.breakpoint
                        count += 1

            print(f"{count} Breakpoints Set: {ctxs.semantic}")



"""
Math Utilities
TODO: Refactor into submodules
"""
import IPython
import logging as root_logger
from functools import partial
from math import atan2, copysign
import numpy as np
from numpy import cos, sin, pi
from scipy.interpolate import splprep, splev

from .constants import PI, TWOPI, QUARTERPI, EPSILON, TOLERANCE
from .constants import IntersectEnum, DELTA, HALFDELTA, NODE_RECIPROCAL
from .constants import SAMPLE_DATA_LEN, LINE_DATA_LEN, BEZIER_DATA_LEN, CIRCLE_DATA_LEN


logging = root_logger.getLogger(__name__)

def construct_matrix_multiplier(a):
    """ Partial Matrix Mul constructor for use in rotate point
    for slices of a 2d array: """
    assert(isinstance(a, np.ndarray))
    return partial(lambda x, y: x @ y, a)

#------------------------------
# def circle functions
#------------------------------
def displace_around_circle(xys, scale, n, random=None):
    """ displace the data around a scaled noisy circle """
    #pylint: disable=invalid-name
    #Create a circle:
    if random is None:
        random = np.random.random
    t = np.linspace(0, 2*pi, n)
    rotation = np.column_stack((sin(t), cos(t))).transpose()
    #create some noise:
    noise = random(n)
    #jitter the rotation:
    jittered = (rotation * noise)
    #control the amount of this noise to apply
    scaled = (jittered * scale).transpose()
    #apply the noise to the data
    mod_points = xys + scaled
    return mod_points

def sample_circle(xyrs, n, sort_rads=True, sort_radi=True, easing=None, random=None):
    """
    https://stackoverflow.com/questions/5837572
    """
    if random is None:
        random = np.random.random

    xyrs_r = xyrs.reshape((-1, 1, CIRCLE_DATA_LEN)).repeat(n, axis=1)

    r = scale_ndarray(np.sqrt(random((n, 1))), xyrs_r[0,:,4:])
    if sort_radi:
        r.sort(axis=1)

    theta = scale_ndarray(random((n, 1)), xyrs_r[0,:,2:4])
    if sort_rads:
        theta.sort(axis=1)

    r_shaped = r.reshape((xyrs_r.shape[0],n,1))
    theta_shaped = theta.reshape((xyrs_r.shape[0],n,1))

    rot = np.dstack((np.cos(theta_shaped), np.sin(theta_shaped)))
    r_x_rot = r_shaped * rot

    result = xyrs_r[:,:,:2] + r_x_rot
    flattened = result.reshape((-1, 2))
    return flattened

def get_circle_3p(p1, p2, p3, arb_intersect=20000):
    """
    Given 3 points, treat them as defining two chords on a circle,
    intersect them to find the centre, then calculate the radius
    Thus: circumcircle
    """
    #pylint: disable=too-many-locals
    assert(all([isinstance(x, np.ndarray) for x in [p1, p2, p3]]))
    sorted_points = sort_coords(np.array([p1, p2, p3]))
    p1 = sorted_points[0]
    p2 = sorted_points[1]
    p3 = sorted_points[2]

    arb_height = arb_intersect
    #mid points and norms:
    m1 = get_midpoint(p1, p2)
    n1 = get_bisector(m1, p2)
    m2 = get_midpoint(p2, p3)
    n2 = get_bisector(m2, p3)
    #extended norms:
    v1 = m1 + (1 * arb_height * n1)
    v2 = m2 + (1 * arb_height * n2)
    v1_i = m1 + (-1 * arb_height * n1)
    v2_i = m2 + (-1 * arb_height * n2)
    #resulting lines:
    l1 = np.row_stack((m1, v1))
    l2 = np.row_stack((m2, v2))
    l1_i = np.row_stack((m1, v1_i))
    l2_i = np.row_stack((m2, v2_i))
    #intersect extended norms:
    #in the four combinations of directions
    i_1 = intersect(l1, l2)
    i_2 = intersect(l1_i, l2_i)
    i_3 = intersect(l1, l2_i)
    i_4 = intersect(l1_i, l2)
    #get the intersection:
    the_intersect = [x for x in [i_1, i_2, i_3, i_4] if x is not None]
    if the_intersect is None or not bool(the_intersect):
        return None
    r1 = get_distance(p1, the_intersect[0])
    r2 = get_distance(p2, the_intersect[0])
    r3 = get_distance(p3, the_intersect[0])

    #a circle only if they are have the same radius
    if np.isclose(r1, r2) and np.isclose(r2, r3):
        return (the_intersect[0], r1)
    else:
        return None

def get_lowest_point_on_circle(centre, radius):
    """ given the centre of a circle and a radius, get the lowest y point on that circle """
    #return centre + np.array([np.cos(THREEFOURTHSTWOPI) * radius,
    #                          np.sin(THREEFOURTHSTWOPI) * radius])
    return centre - np.array([0, radius])

def in_circle(centre, radius, points):
    """ Test a set of points to see if they are within a circle's radius """
    d = get_distance_raw(centre, points)
    return d < pow(radius, 2)


#------------------------------
# def interpolation functions
#------------------------------
def granulate(xys, grains=10, mult=2, random=None):
    """ Given a set of points, offset each slightly
    by the direction between the points
    """
    assert(isinstance(xys, np.ndarray))
    assert(len(xys.shape) == 2)
    if random is None:
        random = np.random.random
    directions, hypos = get_directions(xys)
    granulated = np.zeros((1, 2))
    for i, d in enumerate(hypos):
        sub_granules = xys[i, :] + (d * directions[i, :]*(random((grains, 1))) * mult)
        granulated = np.row_stack((granulated, sub_granules))
    return granulated[1:]

def vary(xys, step_size, pix, random=None):
    """
    FIXME : investigate
    for a given set of points, wiggle them slightly
    """
    assert(isinstance(xys, np.ndarray))
    assert(len(xys.shape) == 2)
    if random is None:
        random = np.random.random
    r = (1.0-2.0 * random((len(xys), 1)))
    scale = np.reshape(np.arange(len(xys)).astype('float'), (len(xys), 1))
    noise = (r* scale * step_size)
    a = random(len(xys))
    rnd = np.column_stack((np.cos(a), np.sin(a)))
    rnd_noise = rnd * noise
    rnd_noise_pix = rnd_noise * pix
    xys_prime = xys + rnd_noise_pix
    return xys_prime

def _interpolate(xy, num_points, smoothing=0.2):
    """ given a set of points, generate values between those points """
    assert(isinstance(xy, np.ndarray))
    assert(len(xy.shape) == 2)
    assert(xy.shape[0] >= 4)
    assert(xy.shape[1] == 2)
    spline_tuple, _ = splprep([xy[:, 0], xy[:, 1]], s=smoothing)
    interpolate_points = np.linspace(0, 1, num_points)
    smoothed_xy = np.column_stack(splev(interpolate_points, spline_tuple))
    return smoothed_xy

def scale(xs, minmax):
    result = minmax[0] + (xs * (minmax[1] - minmax[0]))
    return result

def scale_ndarray(xs, minmaxs):
    mins = minmaxs[:,0].reshape((-1,1))
    ranges = (minmaxs[:,1] - minmaxs[:,0]).reshape((-1,1))
    result = mins + (xs * ranges)
    return result


#------------------------------
# def direction functions
#------------------------------
def get_random_directions(n=1, random=None):
    """ Choose a direction of cardinal and intercardinal directions """
    dirs = [-1, 0, 1]
    if random is None:
        random = lambda a, x: np.random.choice(a, size=x*2, replace=True, p=None)
    result = random(dirs, n).reshape((n, 2))
    return result

def get_directions(xys):
    """ Given a set of points, get the unit direction
    from each point to the next point
    """
    assert(isinstance(xys, np.ndarray))
    assert(len(xys.shape) == 2)
    #convert to vectors:
    #xysPrime.shape = (n, 4)
    #Leading point first to prevent wrap deformation
    # xys_prime = np.column_stack((xys[1:, :], xys[:-1, :]))
    # dx = xys_prime[:, 2] - xys_prime[:, 0]
    # dy = xys_prime[:, 3] - xys_prime[:, 1]
    ds = xys[:-1] - xys[1:]
    #radians:
    arc = np.arctan2(ds[:,1], ds[:,0])
    directions = np.column_stack([np.cos(arc), np.sin(arc)])
    #hypotenuse
    hypos = np.sqrt(np.square(ds[:,0])+np.square(ds[:,1]))
    return np.column_stack((directions, hypos))


#------------------------------
# def line functions
#------------------------------
def intersect(line_1, line_2, tolerance=TOLERANCE):
    """ Get the intersection points of two line segments
    see: http://ericleong.me/research/circle-line/
    so line_1:(start, end), line_2:(start, end)
    returns np.array([x, y]) of intersection or None
    """
    #pylint: disable=too-many-locals
    assert(isinstance(line_1, np.ndarray))
    assert(isinstance(line_2, np.ndarray))
    assert(line_1.shape == (2, 2))
    assert(line_2.shape == (2, 2))
    #The points
    p0 = line_1[0]
    p1 = line_1[1]
    p2 = line_2[0]
    p3 = line_2[1]

    a1 = p1[1] - p0[1]
    b1 = p0[0] - p1[0]
    c1b = a1 * p0[0] + b1 * p0[1]

    a2 = p3[1] - p2[1]
    b2 = p2[0] - p3[0]
    c2b = a2*p2[0] + b2*p2[1]

    detb = a1 * b2 - a2 * b1
    if detb == 0:
        return None

    xb = ((c1b * b2) - (b1 * c2b)) / detb
    yb = ((a1 * c2b) - (c1b * a2)) / detb
    xyb = np.array([xb, yb])

    l1mins = np.min((p0, p1), axis=0) - tolerance
    l2mins = np.min((p2, p3), axis=0) - tolerance
    l1maxs = np.max((p0, p1), axis=0) + tolerance
    l2maxs = np.max((p2, p3), axis=0) + tolerance

    if (l1mins <= xyb).all() and (l2mins <= xyb).all() and \
       (xyb <= l1maxs).all() and (xyb <= l2maxs).all():
        return xyb
    return None

def get_unit_vector(p1, p2):
    """ Given two points, get the normalized direction """
    assert(isinstance(p1, np.ndarray))
    assert(isinstance(p2, np.ndarray))
    d = get_distance(p1, p2)
    if np.allclose(d, 0):
        return np.array([0, 0])
    n = (p2-p1)
    normalized = n / d
    return normalized

def extend_line(p1, p2, m=1, from_start=True):
    """ Extend a line by m units
    Returns the new end points only
    """
    n = get_unit_vector(p1, p2)
    if from_start:
        el = p1 + (n * m)
    else:
        el = p2 + (n * m)
    return el

def is_point_on_line(p, l):
    """ Test to see if a point is on a line """
    assert(isinstance(p, np.ndarray))
    assert(isinstance(l, np.ndarray))
    points = p.reshape((-1, 2))
    the_lines = l.reshape((-1, 2, 2))
    l_mins = the_lines.min(axis=1)
    l_maxs = the_lines.max(axis=1)

    in_bounds_xs = l_mins[:, 0] <= points[:, 0] <= l_maxs[:, 0]
    in_bounds_ys = l_mins[:, 1] <= points[:, 1] <= l_maxs[:, 1]

    if np.allclose((the_lines[:, 0, 0] - the_lines[:, 1, 0]), 0):
        return in_bounds_ys and in_bounds_xs
    slopes = (the_lines[:, 0, 1] - the_lines[:, 1, 1]) / (the_lines[:, 0, 0] - the_lines[:, 1, 0])
    y_intersects = - slopes * the_lines[:, 0, 0] + the_lines[:, 0, 1]
    line_ys = slopes * points[:, 0] + y_intersects
    return np.allclose(line_ys, points[0, 1]) and in_bounds_ys and in_bounds_xs

def make_horizontal_lines(n=1, random=None):
    """ Utility to Describe a horizontal line as a vector of start and end points  """
    if random is None:
        random = np.random.random
    x = random((n, 2)).sort()
    y = random(n).reshape((-1, 1))
    return np.column_stack((x[:, 0], y, x[:, 1], y))

def make_vertical_lines(n=1, random=None):
    """ utility Describe a vertical line as a vector of start and end points """
    if random is None:
        random =np.random.random
    x = random(n).reshape((-1, 1))
    y = random((n, 2)).sort()
    return np.column_stack((x, y[:, 0], x, y[:, 1]))

def sample_along_lines(xys, n, easing=None, override=None):
    """ For a set of lines, sample along them n times,
    with easings distribution. can be hard overriden with override
    """
    t = np.linspace(0,1,n)
    if override is not None:
        t = override
    if easing is not None:
        t = easing(t)
    s_points = t
    s_invert = (1 - s_points)
    num_points = s_points.shape[0]
    fade = np.vstack((s_invert,s_points))
    lines = xys.reshape((-1, 2, 2))
    xs = lines[:, :, 0]
    ys = lines[:, :, 1]
    xsr = xs.repeat(num_points).reshape((-1, 2, num_points))
    ysr = ys.repeat(num_points).reshape((-1, 2, num_points))
    xsrf = xsr * fade
    ysrf = ysr * fade
    s_xs = xsrf.sum(axis=1)
    s_ys = ysrf.sum(axis=1)
    paired = np.hstack((s_xs, s_ys))
    reshaped = paired.reshape((-1, num_points, 2), order='F').reshape((-1,2))
    return reshaped

def create_line(xys, n, easing=None):
    """ Given a start and end, create t number of points along that line """
    line = sample_along_lines(xys.reshape((-1,LINE_DATA_LEN)), n, easing=easing)
    return line

def bezier1cp(a_cp_b, t, f=None, p=None):
    """ Given the start, end, and a control point, create t number of points along that bezier
    t : the number of points to linearly create used to sample along the bezier
    f : a transform function for the sample points prior to calculate bezier
    p : an overriding set of arbitrary sample points for calculate bezier
    """
    #pylint: disable=unused-variable
    assert(isinstance(a_cp_b, np.ndarray))
    if p is not None:
        assert(isinstance(p, np.ndarray))
        sample_points = p
    else:
        sample_points = np.linspace(0, 1, t)
        if f is not None:
            assert(callable(f))
            #f is an easing lookup function
            sample_points = f(t)

    len_samples = len(sample_points)
    quadratic_matrix = np.array([[[1,0,0],[-2,2,0],[1,-2,1]]])
    t_matrix = lambda x: np.array([1, x,  pow(x,2)])

    points = a_cp_b.reshape((-1,3,2))
    results = np.zeros((len(a_cp_b),1,2))
    for t in sample_points:
        t_mul_mat = t_matrix(t) @ quadratic_matrix
        t_lines = t_mul_mat @ points
        results = np.hstack((results, t_lines.reshape(-1,1,2)))
    return results[:,1:,:].reshape((-1,2))

def bezier2cp(a_cpcp_b, n=None, p=None, easing=None):
    """ Given a start, end, and two control points along the way,
    create n number of points along that bezier
    n : The number of points to sample linearly
    f : the transform function for the linear sampling
    p : arbitrary points to use for sampling instead
    Matrix Multiplication algorithm from:
    https://pomax.github.io/bezierinfo/#decasteljau
    """
    #pylint: disable=too-many-locals
    #pylint: disable=unused-variable
    assert(isinstance(a_cpcp_b, np.ndarray))
    if p is not None:
        assert(isinstance(p, np.ndarray))
        sample_points = p
    elif n is not None:
        sample_points = np.linspace(0, 1, n)
    else:
        raise Exception("Neither arbitrary points or n given")
    if easing is not None:
        assert(callable(easing))
        sample_points = easing(sample_points)

    len_samples = len(sample_points)
    cubic_matrix = np.array([[1,0,0,0],[-3,3,0,0],[3,-6,3,0],[-1,3,-3,1]])
    t_matrix = lambda x: np.array([1, x,  pow(x,2), pow(x,3)])

    points = a_cpcp_b.reshape((-1,4,2))
    results = np.zeros((len(a_cpcp_b),1,2))
    for t in sample_points:
        t_mul_mat = t_matrix(t) @ cubic_matrix
        t_lines = t_mul_mat @ points
        results = np.hstack((results, t_lines.reshape(-1,1,2)))
    return results[:,1:,:].reshape((-1,2))


#------------------------------
# def distance functions
#------------------------------
def check_distance_from_points(point, quad_tree, dist=HALFDELTA):
    """ Given a point and a quadtree, return true if the point is within
    the bounds of the quadtree but not near any points """
    bbox = [point[0] - dist, point[1] - dist, point[0] + dist, point[1] + dist]
    area = quad_tree.intersect(bbox)
    inside_canvas = point[0] > 0 and point[0] < 1.0 and point[1] > 0 and point[1] < 1.0
    return (not bool(area)) and inside_canvas

def get_closest_to_focus(focus, possible_points):
    """ Given a set of points, return the point closest to the focus """
    ds = get_distance(focus, possible_points)
    m_d = ds.min()
    i = ds.tolist().index(m_d)
    return possible_points[i]

def get_closest_on_side(ref_point, possible_points, left=True):
    """
    given a reference point and a set of candidates, get the closest
    point on either the left or right of that reference
    """
    subbed = possible_points - ref_point
    if left:
        on_side = subbed[:, 0] < 0
    else:
        on_side = subbed[:, 0] > 0
    try:
        i = on_side.tolist().index(True)
        return possible_points[i]
    except ValueError:
        return None

def get_distance_raw(p1, p2):
    """ Get the non-square-root distance for pairs of points """
    assert(isinstance(p1, np.ndarray))
    assert(isinstance(p2, np.ndarray))
    p1 = p1.reshape(-1, 2)
    p2 = p2.reshape(-1, 2)
    d_squared = pow(p2-p1, 2)
    #summed = dSquared[:, 0] + dSquared[:, 1]
    summed = d_squared.sum(axis=1)
    return summed

def get_distance(p1, p2):
    """ Get the square-root distance of pairs of points """
    assert(isinstance(p1, np.ndarray))
    assert(isinstance(p2, np.ndarray))

    summed = get_distance_raw(p1, p2)
    sqrtd = np.sqrt(summed)
    return sqrtd

def get_distance_xyxy(x1, y1, x2, y2):
    """ Utility to get the raw distance of points as separate x's and y's  """
    return get_distance_raw(np.array([x1, y1]), np.array([x2, y2]))[0]

def get_midpoint(p1, p2):
    """ Given two points, get the point directly between them """
    m = (p1 + p2) / 2
    return m


#------------------------------
# def rotate functions
#------------------------------
def radians_between_points(a, b):
    """ takes np.arrays
        return the radian relation of b to a (source)
        ie: if > 0: anti-clockwise, < 0: clockwise
    """
    c = b - a
    return atan2(c[1], c[0])

def is_clockwise(*args, cartesian=True):
    """ Test whether a set of points are in clockwise order  """
    #based on stackoverflow.
    #sum over edges, if positive: CW. negative: CCW
    #assumes normal cartesian of y bottom = 0
    the_sum = 0
    p1s = args
    p2s = list(args[1:])
    p2s.append(args[0])
    pairs = zip(p1s, p2s)
    for p1, p2 in pairs:
        a = (p2[0]-p1[0]) * (p2[1]+p1[1])
        the_sum += a
    if cartesian:
        return the_sum >= 0
    else:
        return the_sum < 0

def is_counter_clockwise(a, b, c):
    """ Given 3 points, do they form a counter clockwise turn """
    assert(all([isinstance(x, np.ndarray) for x in [a, b, c]]))
    offset_b = b - a
    offset_c = c - a
    crossed = np.cross(offset_b, offset_c)
    return crossed >= 0

def get_bisector(p1, p2, r=False):
    """ With a normalised line, rotate 90 degrees,
    r=True : to the right
    r=False : to the left
    """
    n = get_unit_vector(p1, p2)
    if r:
        n_prime = n.dot([[0, -1],
                         [1, 0]])
    else:
        n_prime = n.dot([[0, 1],
                         [-1, 0]])
    return n_prime

def rotate_point(p, cen=None, rads=None, rad_min=-QUARTERPI, rad_max=QUARTERPI, random=None):
    """ Given a point, rotate it around a centre point by either radians,
    or within a range of radians
    """
    #p1 = cen, p2=point, @ is matrix mul
    if cen is None:
        cen = np.array([0, 0])
    if rads is None:
        use_radians = random_radian(min_v=rad_min, max_v=rad_max, random=random)
        if isinstance(use_rads, np.ndarray):
            use_radians = use_rads[0]
    else:
        use_radians = rads
    #apply to 1d slices, this allows multiple points to be
    #passed into the function together,
    #without messing up the rotation matmul
    rot_m = rotation_matrix(use_radians)
    offset = (p - cen)
    if len(p.shape) == 1:
        applied = rot_m @ offset
    else:
        applied = np.apply_along_axis(construct_matrix_multiplier(rot_m), 1, offset)
    result = cen + applied
    return result

def __rotate_point_obsolete(p, cen, rads):
    """ Does what rotate point does, explicitly instead
    of with matrix multiplication """
    assert(len(p.shape) == 2)
    c = np.cos(rads)
    s = np.sin(rads)
    centred = p - cen
    cos_p = centred * c
    sin_p = centred * s
    nx = cos_p[:, 0] - sin_p[:, 1]
    ny = sin_p[:, 0] + cos_p[:, 1]
    un_centered = np.column_stack((nx, ny)) + cen
    return un_centered

def random_radian(min_v=-TWOPI, max_v=TWOPI, shape=(1,), random=None):
    """ Get a random value within the range of radians -2pi -> 2pi """
    if random is None:
        random = np.random.random
    return min_v + (random(shape) * (max_v-min_v))

def rotation_matrix(rad):
    """ Get a matrix for rotating a point by an amount of radians """
    return np.array([[cos(rad), -sin(rad)],
                     [sin(rad), cos(rad)]])


#------------------------------
# def point functions
#------------------------------
def node_to_position(x, y):
    """ Convert a nodes XY Position to its actual position """
    return [NODE_RECIPROCAL * x, NODE_RECIPROCAL * y]

def calculate_single_point(points, d=DELTA, random=None):
    """ points passed in, move in a random direction """
    if random is None:
        random = np.random.random
    arr = random((points.shape[0], 2)) * TWOPI
    delta = np.array([sin(arr[:, 0]), cos(arr[:, 1])]) * (2 * d)
    return points + delta

def calculate_vector_point(ps, d=DELTA, random=None):
    """ passed in pairs of points, move in the direction of the vector """
    if random is None:
        random = np.random.random
    vector = ps[:, 2:] - ps[:, :2]
    rand_amnt = random((ps.shape[0], 2)) * TWOPI
    mag = np.sqrt(np.sum(np.square(vector)))
    norm_vector = vector / mag
    move_vector = norm_vector * (2 * d)
    jiggled_vector = move_vector * np.array([sin(rand_amnt[:, 0]), cos(rand_amnt[:, 1])])
    return ps[:, 2:] + jiggled_vector

def sort_coords(arr):
    """ Sort a list of points by x then y value  """
    ind = np.lexsort((arr[:, 1], arr[:, 0]))
    return arr[ind]

def random_points(n, random=None):
    """ utility to get n 2d points """
    if random is None:
        random = np.random.random
    return random(n*2)


#------------------------------
# def bbox functions
#------------------------------
def bbox_to_lines(bbox, epsilon=EPSILON):
    """ take in the min and max values of a bbox,
    return back a list of 4 lines with the enum designating their position """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    bbox_e = bbox + np.array([-epsilon, -epsilon, epsilon, epsilon])
    # [[minx, miny], [maxx, maxy]] -> [[minx, maxx], [miny, maxy]]
    bbox_t = bbox_e.reshape((2, 2)).transpose()
    #convert the bbox to bounding lines
    select_x = np.array([1, 0])
    select_y = np.array([0, 1])
    mins = bbox_t[:, 0]
    maxs = bbox_t[:, 1]
    min_x_max_y = mins * select_x + maxs * select_y
    max_x_min_y = maxs * select_x + mins * select_y
    lines = [(np.row_stack((mins, max_x_min_y)), IntersectEnum.HBOTTOM),
             (np.row_stack((min_x_max_y, maxs)), IntersectEnum.HTOP),
             (np.row_stack((mins, min_x_max_y)), IntersectEnum.VLEFT),
             (np.row_stack((max_x_min_y, maxs)), IntersectEnum.VRIGHT)]

    return lines

def bound_line_in_bbox(line, bbox):
    """ takes in a line, limits it to be within a bbox """
    #replace original line endpoint with intersection point
    bbl = bbox_to_lines(bbox)
    intersections = [x for x in [intersect(line, x) for x, y in bbl] if x is not None]
    if not bool(intersections):
        return [line]
    return [np.array([line[0], x]) for x in intersections]

def calc_bbox_corner(bbox, ies, epsilon=EPSILON):
    """ Calculate the nearest corner of a bbox for set of existing intersections  """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    assert(isinstance(ies, set))
    hb = IntersectEnum.HBOTTOM
    ht = IntersectEnum.HTOP
    vl = IntersectEnum.VLEFT
    vr = IntersectEnum.VRIGHT
    bbox_e = bbox + np.array([-epsilon, -epsilon, epsilon, epsilon])
    # [[minx, miny], [maxx, maxy]] -> [[minx, maxx], [miny, maxy]]
    bbox_t = bbox_e.reshape((2, 2)).transpose()
    #convert the bbox to bounding lines
    select_x = np.array([1, 0])
    select_y = np.array([0, 1])
    mins = bbox_t[:, 0]
    maxs = bbox_t[:, 1]
    min_x_max_y = mins * select_x + maxs * select_y
    max_x_min_y = maxs * select_x + mins * select_y

    if ies.issubset([hb, vl]):
        return mins
    elif ies.issubset([hb, vr]):
        return max_x_min_y
    elif ies.issubset([ht, vl]):
        return min_x_max_y
    elif ies.issubset([ht, vr]):
        return maxs
    else:
        raise Exception("Calculating box corner failed for: {}".format(ies))

def within_bbox(point, bbox, tolerance=TOLERANCE):
    """ Test whether a point is within the given bbox """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    assert(isinstance(point, np.ndarray))
    assert(point.shape == (2, ))
    mod_bbox = bbox + np.array([-tolerance, -tolerance, tolerance, tolerance])
    in_x_bounds = mod_bbox[0] < point[0] and point[0] < mod_bbox[2]
    in_y_bounds = mod_bbox[1] < point[1] and point[1] < mod_bbox[3]
    return in_x_bounds and in_y_bounds

def bbox_centre(bbox):
    """ Get the centre of a bbox """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    bbox_t = bbox.reshape((2, 2)).transpose()
    mins = bbox_t[:, 0]
    maxs = bbox_t[:, 1]
    ranges = maxs - mins
    mid = ranges * 0.5
    return mid

def make_bbox_from_point(point, i):
    """ Given a centre point and a length, create a bbox """
    border = HALFDELTA * (1/(i+1))
    return [point[0] - border, point[1] - border, point[0] + border, point[1] + border]

def check_sign(a, b):
    """ Test whether two numbers have the same sign """
    return copysign(a, b) == a

def get_min_range_pair(p1, p2):
    """ TODO: Can't remember, test this """
    d1 = get_distance(p1, p2)
    fp2 = np.flipud(p2)
    d2 = get_distance(p1, fp2)
    d1_min = d1.min()
    d2_min = d2.min()
    if d1_min < d2_min:
        i = d1.tolist().index(d1_min)
        #get the right xs
        return np.array([p1[i][0], p2[i][0]])
    else:
        i = d2.tolist().index(d2_min)
        return np.array([p1[i][0], fp2[i][0]])

def clamp(n, minn=0, maxn=1):
    """ Clamp a number between min and max,
    could be replaced with np.clip
    """
    return max(min(maxn, n), minn)

def get_ranges(a):
    """ Given pairs, get the ranges of them """
    assert(isinstance(a, np.ndarray))
    assert(a.shape[1] == 2)
    ranges = np.array([a.min(axis=0), a.max(axis=0)])
    return ranges.T


#------------------------------
# def SAMPLING WRAPPER
#------------------------------
def sample_wrapper(func, data, n, radius, colour, easing=None, random=None):
    """ A Wrapper for sample_circle, sample_along_lines,
    bezier1cp and bezier2cp
    Takes the function, applies the data and samp_sig_or_count to it,
    then duplicates the radius and colours to the appropriate count,
    and combines
    """
    assert(callable(func))
    sampled = func(data, n, easing=easing, random=random)
    radius_size = len(data) * n
    repeated_radius = np.repeat(radius, radius_size).reshape((radius_size, 1))
    repeated_colours = np.repeat(colour.reshape((-1,4)), len(sampled), axis=0)
    combined = np.column_stack((sampled, repeated_radius, repeated_colours))
    return combined


#------------------------------
# def DEPRECATED
#------------------------------
def get_normal(p1, p2):
    """ Get the normalized direction from two points """
    raise Exception("Deprecated: Use get_unit_vector")

def angle_between_points(a, b):
    raise Exception("Deprecated: Use radians_between_points")
